'''
Utilities for generating "archetecture independant" ASTs.
'''
from vivisect.const import *
from vivisect.symboliks.common import *

def wipeAstArch(symctx, symobjs, emu=None, wipeva=False):
    '''
    Given a symbolik analysis context, modify a set of
    symbolik states to be "arch independant" while
    maintaining register context awareness in a "portable"
    way...

    Goals:

        Given:
          ( arg0 + ecx ) * ( ebx + ebx )

        modify the AST such that register selection and
        order do *not* effect the solve values such that:

        == ( arg0 + ebx ) * ( ecx + ecx )
        != ( arg0 + ecx ) * ( ebx + ecx )
        != ( arg0 + ecx ) * ( ebx + eax )

        *while* maintaining the order independance of all
        appropriate operators.

        Result:
          ( arg0 + reg0 ) * ( reg1 + reg1 )

          NOTE: the key here is that "order" of regN generation
                is stable across symbolik states generated with
                different "order of occurance" of register access.

          ( ie, still == ( edx + edx ) * ( ecx + arg0 ) )

    Specify wipeva=True to similarly "replace" all virtual addresses
    within the AST.
    '''
    archemu = symctx.vw.getCachedEmu('sym:archind')

    # step one, create symid->reg mapping and replace all
    # sym vars that are regs with a constant value

    # for constants to map new symobj id -> oldsym
    idtova = {}
    # for registers to map new symobj id -> oldsym
    idtoold = {}

    # a tree walker to frob reg vars
    def normast(path, oldsym, ctx):
        # are we wipping away consts?
        if wipeva and oldsym.symtype == SYMT_CONST:
            if symctx.vw.isValidPointer(oldsym.value):
                # check for function thunks
                if symctx.vw.isFunction(oldsym.value):
                    api = symctx.vw.getFunctionMeta(oldsym.value, 'Thunk')
                    if api:
                        return Var(api, width=symctx.vw.psize)

                newobj = Var('piva_global',4)
                idtova[newobj._sym_id] = oldsym
                return newobj

        if oldsym.symtype != SYMT_VAR:
            return None

        # check if this is a register
        if archemu.getRegisterIndex(oldsym.name) is None:
            return None

        # do *not* change the var name below.  it's value is
        # intrinsic to the comparibility of results over time
        newobj = Var('archindreg', 4)  # norm all to 4...

        idtoold[newobj._sym_id] = oldsym
        return newobj

    # frob regs
    symobjs = [s.walkTree(normast) for s in symobjs]
    # force the solver cache to populate
    [s.solve(emu=emu) for s in symobjs]
    # retrieve all "position hashes" or whatever...
    vapos = []
    sympos = []

    def gatherpos(path, symobj, ctx):
        if idtova.get(symobj._sym_id) is not None:
            poshash = 'va:' + (':'.join(['%.8x' % s.solve(emu=emu) for s in path]))
            vapos.append((poshash, symobj._sym_id))
            return

        if idtoold.get(symobj._sym_id) is not None:
            poshash = 'sym:' + (':'.join(['%.8x' % s.solve(emu=emu) for s in path]))
            sympos.append((poshash, symobj._sym_id))
            return

    [s.walkTree(gatherpos) for s in symobjs]

    # sort them by the "order stabilized" "position hashes"
    vapos.sort()
    sympos.sort()

    # generate the variable mapping
    indregs = {}
    for poshash, symid in sympos:
        origreg = idtoold.get(symid).name
        indreg = indregs.get(origreg)
        if indreg is None:
            # do *not* modify this register name convention
            # (all previously calcualated results are only
            # comparable while these are the same. Additionally,
            # having the number "early" in the string, helps
            # spread the entropy further... (python hash)
            indreg = '%dindreg' % len(indregs)
            indregs[origreg] = indreg

    indvas = {}
    # Narchindva
    basename = "%darchindva"
    idx = 0
    for poshash, symid in vapos:
        origva = idtova.get(symid).value
        indva = indvas.get(origva)
        if indva is None:
            indva = basename % idx
            indvas[origva] = indva
            idx += 1

    # re-build the tree with the new "ind reg" vars
    def makeind(path, symobj, ctx):
        oldsym = idtoold.get(symobj._sym_id)
        if oldsym is not None:
            return Var(indregs.get(oldsym.name), 4)

        oldsym = idtova.get(symobj._sym_id)
        if oldsym is not None:
            obj = Var(indvas.get(oldsym.value), 4)
            ploc = symctx.vw.getLocation(oldsym.value)
            if ploc:
                va, size, ltype, tinfo = ploc
                if ltype == LOC_STRING or ltype == LOC_UNI:
                    obj._str_repr = symctx.vw.reprLocation(ploc)
            return obj
    symobjs = [s.walkTree(makeind) for s in symobjs]
    return symobjs



import vdb.extensions.i386 as v_ext_i386

import vdb.extensions.i386 as vdb_ext_i386

def vdbExtension(vdb, trace):
    vdb.addCmdAlias('db','mem -F bytes')
    vdb.addCmdAlias('dw','mem -F u_int_16')
    vdb.addCmdAlias('dd','mem -F u_int_32')
    vdb.addCmdAlias('dq','mem -F u_int_64')
    vdb.addCmdAlias('dr','mem -F "Deref View"')
    vdb.addCmdAlias('ds','mem -F "Symbols View"')

    vdb.registerCmdExtension(vdb_ext_i386.eflags,'amd64')



'''
Perform further analysis on 32-bit GO language executables.

This module extends code flow analysis to runtime_main for Windows PE binaries.
GO binaries start from a single export and proceed thru several functions
that initialize GO.  Specific application code is launched from the GO function
runtime_main(), which is invoked by "call eax" with an address placed on the
stack many calls earlier.  Vivisect code flow analysis does not track the
address; this module finds the address and invokes makeFunction(va).

Samples have been found with different instruction sequences for reaching the
address of runtime_main(); this module attempts all observed sequences.
'''

import logging

import envi
import envi.archs.i386.disasm

logger = logging.getLogger(__name__)

# Opcode sequence, where element [-5] is the important one.
_GOLANG_I386_INSTRS = ['cld', 'call', 'mov', 'mov', 'mov', 'mov', 'call',
                       'call', 'call', 'push', 'push', 'call', 'pop', 'pop']

def analyze(vw):
    '''
    Perform further analysis on GO language executables.
    '''
    # Make sure it is a PE file, with "Go build ID:" in the first few bytes
    # of the .text segment (versus a .upxN segment of a packed sample).
    has_go_build = False
    for segment in vw.getSegments():
        va, size, name, filename = segment
        if name == '.text':
            bytez = vw.readMemory(va, 10000)
            k1 = bytez.find(b'Go build ID: ')
            if k1 != -1:
                has_go_build = True
                break
    if not has_go_build:
        return

    # There should be just one entry point (the public export).
    ep = vw.getEntryPoints()
    if len(ep) != 1:
        return

    # Search from the entry point function.
    ep_va = ep[0]
    bblocks = vw.getFunctionBlocks(ep_va)
    if not bblocks:
        return
    ptr_va, runtime_va = extract_golang_mainmain(vw, bblocks, filename)
    if runtime_va is None:
        return

    # Invoke codeflow on runtime_main().
    vw.addEntryPoint(runtime_va)
    logger.debug('discovered runtime function: 0x%x', runtime_va)
    vw.makeFunction(runtime_va)

    # Also mark the ptr_va as a pointer to runtime_va.
    vw.makePointer(ptr_va, tova=runtime_va)


def extract_golang_mainmain(vw, basic_blocks, filename):
    '''
    Find the basic block of interest and return the address of
    the pointer and its contents, runtime_main().

    The BB will contain the sequence of opcodes in _GOLANG_I386_INSTRS.
    The push instruction -5 from the end will load the contents of a memory
    address, and those contents are the runtime_main() address.
    '''
    op = find_golang_bblock(vw, basic_blocks, _GOLANG_I386_INSTRS, 5)
    if op is None:
        op = find_golang_bblock_via_stack(vw, filename)
        if op is None:
            return None, None

    # The key opcode is "push immediate", which identifies a pointer which
    # in turn identifies the GO function runtime_mainPC (aka runtime_main).
    ptr_va, runtime_va = parse_push_imm(vw, op, filename, get_content=True)
    return ptr_va, runtime_va


def find_golang_bblock(vw, basic_blocks, match, idx):
    '''
    Find the basic block of interest and return the opcode where
    the special sequence of opcodes begins.  Return None if not found.

    For this case, Vivisect interprets the entry as one large function
    (IDA Pro may break it into 2 or 3 functions linked by jmp instructions).
    The basic block of interest is in this function.
    '''
    the_bblock = None
    for bblock in basic_blocks:
        next_va = bblock[0]
        instrs = []
        while next_va < bblock[0] + bblock[1]:
            try:
                op = vw.parseOpcode(next_va)
            except envi.InvalidInstruction:
                op = None
            if op is None:
                return None
            if (len(instrs) > 0) or (op.mnem == match[0]):
                # Record instructions from the first opcode match.
                instrs.append(op)
            next_va += op.size
        if len(instrs) >= len(match):
            the_bblock = bblock
            break
    if the_bblock is None:
        return None

    for k in range(len(match)):
        if instrs[k].mnem != match[k]:
            return None
    return instrs[len(match) - idx]


def find_golang_bblock_via_stack(vw, filename):
    '''
    Find the basic block of interest and return the address where
    the special sequence of opcodes begins.  Return None if not found.

    Some GO executables use a stack manipulation in the entry point,
    and special logic is needed to locate the BB of interest.
    Example:  1897d2de0837090ec350004ef2f9fa87.
    '''
    # Start from the entry point (we already know there is only one).
    ep_va = vw.getEntryPoints()[0]

    # Expect a function with one BB, with a push to the next function
    # just before returning.
    basic_blocks = vw.getFunctionBlocks(ep_va)
    if len(basic_blocks) != 1:
        return None
    instrs = golang_collect_opcodes(vw, basic_blocks[0])
    if (len(instrs) < 2) or \
       (instrs[-1].mnem != 'ret') or \
       (instrs[-2].mnem != 'push'):
        return None
    op = instrs[-2]
    ptr_va, _ = parse_push_imm(vw, op, filename)
    if not ptr_va:
        return None

    # Analyze the function at the pointer if necessary.
    if not vw.isFunction(ptr_va):
        logger.debug('discovered new function(ptr): 0x%x', ptr_va)
        vw.makeFunction(ptr_va)

    # This function should contain the special basic block.
    basic_blocks = vw.getFunctionBlocks(ptr_va)
    return find_golang_bblock(vw, basic_blocks, _GOLANG_I386_INSTRS, 5)


def parse_push_imm(vw, opcode, filename, get_content=False):
    '''
    Parse an opcode that should be "push immediate", returning
    the address of the second operand.  Also return the content of the
    address if so requested.
    Return None, None if there is an error.
    '''
    if len(opcode.opers) != 1:
        return None, None
    if not isinstance(opcode.opers[0], envi.archs.i386.disasm.i386ImmOper):
        return None, None
    try:
        ptr_va = opcode.opers[0].getOperValue(opcode)
        if len(vw.readMemory(ptr_va, 4)) != 4:
            return None, None
        if get_content:
            runtime_va = vw.castPointer(ptr_va)
            if len(vw.readMemory(runtime_va, 4)) != 4:
                return None, None
        else:
            runtime_va = None
        return ptr_va, runtime_va
    except Exception as e:
        return None, None


def golang_collect_opcodes(vw, basic_block):
    '''
    Return a list of opcodes in a basic block.
    Return an empty list of an opcode could not be parsed.
    '''
    next_va = basic_block[0]
    opcodes = []
    while next_va < basic_block[0] + basic_block[1]:
        try:
            op = vw.parseOpcode(next_va)
        except envi.InvalidInstruction:
            return []
        opcodes.append(op)
        next_va += op.size
    return opcodes



import logging
import unittest
import vqt.menubuilder as vmb

logger = logging.getLogger(__name__)


class MenuTest(unittest.TestCase):

    def test_actioncall_success(self):
        ac = vmb.ActionCall(acsucceed, 4, 8)
        out = ac()
        self.assertEqual(out, 32)

    def test_actioncall_fail(self):
        ac = vmb.ActionCall(acfail, 4, 8)
        ac()


def acsucceed(x, y):
    return x * y


def acfail(x, y):
    return y / 0



from vivisect.symboliks.common import *
from vivisect.symboliks.effects import *

class SymbolikTranslator:
    '''
    The SymbolikTranslator is responsible for translating architecture specific
    sequences of Envi Opcode objects into a sequence of purely symbolik effects.
    '''

    def __init__(self, vw):
        self.vw = vw
        self._eff_log = []
        self._con_log = []
        self._op_methods = {}
        for name in dir(self):
            if name.startswith("i_"):
                self._op_methods[name[2:]] = getattr(self, name)
        self._cur_va = None

    def effSetVariable(self, rname, rsym):
        '''
        This may *only* be called with 'natural' register definitions or
        pure symbols...
        (any meta register processing is the responsiblity of the translator
        calling this interface!)
        '''
        self._eff_log.append(SetVariable(self._cur_va, rname, rsym))

    def effReadMemory(self, symaddr, symsize):
        self._eff_log.append(ReadMemory(self._cur_va, symaddr, symsize))
        return Mem(symaddr, symsize)

    def effWriteMemory(self, symaddr, symsize, symobj):
        self._eff_log.append(WriteMemory(self._cur_va, symaddr, symsize, symobj))

    def effFofX(self, funcsym, argsyms=None):
        self._eff_log.append(CallFunction(self._cur_va, funcsym, argsyms))

    def effConstrain(self, addrsym, conssym):
        self._con_log.append(ConstrainPath(self._cur_va, addrsym, conssym))

    def effDebug(self, msg):
        self._eff_log.append(DebugEffect(self._cur_va, msg))

    def translateOpcode(self, op):
        self._cur_va = op.va
        meth = self._op_methods.get(op.mnem, None)
        if meth is None:
            # print('Symboliks: %s: %s Needs: %s' % (hex(op.va), self.__class__.__name__, repr(op)))
            self.effDebug("%s Needs %s" % (self.__class__.__name__, repr(op)))
            return DebugEffect(op.va, "%s Needs %s" % (self.__class__.__name__, repr(op)))

        # instruction translator methods may return branches / constraints
        ret = meth(op)
        if ret is not None:
            for symaddr, symcons in ret:
                self.effConstrain(symaddr, symcons)

        return ret

    def getEffects(self, copy=False):
        '''
        Return the list of symboliks effects which have been logged by
        this translator.
        '''
        if copy:
            return list(self._eff_log)
        return self._eff_log

    def getConstraints(self, copy=False):
        '''
        Return the list of constraints which have been logged by this
        translator.
        '''
        if copy:
            return list(self._con_log)
        return self._con_log

    def clearEffects(self):
        '''
        Clear the translator's list of symbolik effects thus far.
        '''
        self._eff_log = []
        self._con_log = []



import logging
import unittest

import vstruct.defs.macho as vs_macho
import vstruct.defs.macho.const as vsm_const

import vivisect.tests.utils as utils
import vivisect.tests.helpers as helpers
import vivisect.parsers.macho as v_p_macho

logger = logging.getLogger(__name__)


class MachOTests(utils.VivTest):
    @classmethod
    def setUpClass(cls):
        super(MachOTests, cls).setUpClass()
        byts = helpers.getTestBytes('osx', 'amd64', 'safaridriver.bin')
        cls.safari = vs_macho.mach_o()
        cls.safari.vsParse(byts)
        # TODO: plumb a full workspace test after updating the parser

    def test_basic(self):
        header = self.safari.mach_header
        self.eq(vsm_const.MH_MAGIC_64, header.magic)
        self.eq(vsm_const.CPU_TYPE_X86_64, header.cputype)
        self.eq(0x80000003, header.cpusubtype)
        self.eq(vsm_const.MH_EXECUTE, header.filetype)
        self.eq(20, header.ncmds)
        self.eq(1640, header.sizeofcmds)
        self.eq(vsm_const.MH_PIE, header.flags & vsm_const.MH_PIE)
        self.eq(vsm_const.MH_TWOLEVEL, header.flags & vsm_const.MH_TWOLEVEL)
        self.eq(vsm_const.MH_DYLDLINK, header.flags & vsm_const.MH_DYLDLINK)
        self.eq(vsm_const.MH_NOUNDEFS, header.flags & vsm_const.MH_NOUNDEFS)

        self.len(self.safari.load_commands._vs_fields, header.ncmds)

        answers = {
            'cmd0': {
                'cmd': 25,
                'cmdsize': 72,
                'segname': '__PAGEZERO',
                'vmaddr': 0,
                'vmsize': 4294967296,
                'fileoff': 0,
                'filesize': 0,
                'maxprot': 0,
                'initprot': 0,
                'nsects': 0,
                'flags': 0,
            },
            'cmd1': {
                'cmd': 25,
                'cmdsize': 472,
                'segname': '__TEXT',
                'vmaddr': 4294967296,
                'vmsize': 4096,
                'fileoff': 0,
                'filesize': 4096,
                'maxprot': 7,
                'initprot': 5,
                'nsects': 5,
                'flags': 0,
            },
            'cmd2': {
                'cmd': 25,
                'cmdsize': 312,
                'segname': '__DATA',
                'vmaddr': 4294971392,
                'vmsize': 4096,
                'fileoff': 4096,
                'filesize': 4096,
                'maxprot': 7,
                'initprot': 3,
                'nsects': 3,
                'flags': 0,
            },
            'cmd3': {
                'cmd': 25,
                'cmdsize': 72,
                'segname': '__LINKEDIT',
                'vmaddr': 4294975488,
                'vmsize': 12288,
                'fileoff': 8192,
                'filesize': 10128,
                'maxprot': 7,
                'initprot': 1,
                'nsects': 0,
                'flags': 0,
            },
            'cmd4': {
                'cmd': 0x80000022,
                'cmdsize': 48,
                'rebase_off': 8192,
                'rebase_size': 8,
                'bind_off': 0x2008,
                'bind_size': 0x18,
                'weak_bind_off': 0,
                'weak_bind_size': 0,
                'lazy_bind_off': 0x2020,
                'lazy_bind_size': 0x20,
                'export_off': 0x2040,
                'export_size': 0x20,
            },
            'cmd5': {
                'cmd': 2,
                'cmdsize': 0x18,
                'symoff': 0x2068,
                'nsyms': 4,
                'stroff': 0x20b8,
                'strsize': 0x50,
            },
            'cmd6': {
                'cmd': 0xb,
                'cmdsize': 0x50,
                'ilocalsym': 0,
                'nlocalsym': 1,
                'iextdefsym': 1,
                'nextdefsym': 1,
                'iundefsym': 2,
                'nundefsym': 2,
                'tocoff': 0,
                'ntoc': 0,
                'modtaboff': 0,
                'nmodtab': 0,
                'extrefsymoff': 0,
                'nextrefsyms': 0,
                'indirectsymoff': 0x20a8,
                'nindirectsyms': 4,
                'extreloff': 0,
                'nextrel': 0,
                'locreloff': 0,
                'nlocrel': 0,
            },
            'cmd8': {
                'cmd': 0x1b,
                'cmdsize': 24,
                'uuid': "b'\\xbc\\xaaJ\\x0b\\xbfp:]\\xbc\\xf9r\\xf3\\x97\\x80\\xebg'",
            },
            'cmd19': {
                'cmd': 29,
                'cmdsize': 16,
                'dataoff': 8464,
                'datasize': 9856,
            }
        }

        for seg, seginfo in answers.items():
            cmd = self.safari.load_commands[seg]
            self.eq(len(cmd), cmd.cmdsize)
            for key, valu in seginfo.items():
                if type(cmd[key]._vs_value) is bytes:
                    self.eq(str(cmd[key]), valu)
                else:
                    self.eq(cmd[key], valu)

    def test_sections(self):
        answers = {
            'cmd1': [
                {
                    'sectname': '__text',
                    'segname': '__TEXT',
                    'addr': 0x100000a80,
                    'size': 0xa,
                    'offset': 0xa80,
                    'align': 0,
                    'reloff': 0,
                    'nreloc': 0,
                    'flags': 0x80000400,
                    'reserved1': 0,
                    'reserved2': 0,
                    'reserved3': 0,
                },
                {
                    'sectname': '__stubs',
                    'segname': '__TEXT',
                    'addr': 0x100000a8a,
                    'size': 0x6,
                    'offset': 0xa8a,
                    'align': 1,
                    'reloff': 0,
                    'nreloc': 0,
                    'flags': 0x80000408,
                    'reserved1': 0,
                    'reserved2': 6,
                    'reserved3': 0,
                },
                {
                    'sectname': '__stub_helper',
                    'segname': '__TEXT',
                    'addr': 0x100000a90,
                    'size': 0x1a,
                    'offset': 0xa90,
                    'align': 2,
                    'reloff': 0,
                    'nreloc': 0,
                    'flags': 0x80000400,
                    'reserved1': 0,
                    'reserved2': 0,
                    'reserved3': 0,
                },
                {
                    'sectname': '__info_plist',
                    'segname': '__TEXT',
                    'addr': 0x100000aaa,
                    'size': 0x50e,
                    'offset': 0xaaa,
                    'align': 0,
                    'reloff': 0,
                    'nreloc': 0,
                    'flags': 0,
                    'reserved1': 0,
                    'reserved2': 0,
                    'reserved3': 0,
                },
                {
                    'sectname': '__unwind_info',
                    'segname': '__TEXT',
                    'addr': 0x100000fb8,
                    'size': 0x48,
                    'offset': 0xfb8,
                    'align': 2,
                    'reloff': 0,
                    'nreloc': 0,
                    'flags': 0,
                    'reserved1': 0,
                    'reserved2': 0,
                    'reserved3': 0,
                },
            ]
        }

        for seg, sectinfo in answers.items():
            cmd = self.safari.load_commands[seg]
            for idx, section in enumerate(sectinfo):
                realsect = cmd.sections[idx]
                for name, valu in section.items():
                    if type(realsect[name]._vs_value) is bytes:
                        self.eq(valu, str(realsect[name]))
                    else:
                        self.eq(valu, realsect[name])

    def test_version_segment(self):
        for idx, lc in self.safari.load_commands:
            if lc.cmd == vsm_const.LC_SOURCE_VERSION:
                ver = lc.getVersion()
                self.eq('7605.1.33.1.4', ver)






MODE_ARM        = 0
MODE_THUMB      = 1
MODE_JAZELLE    = 2
MODE_THUMBEE    = 3


#Support for different ARM Instruction set versions

# name          bitmask                    decimal         hex
REV_ARMv4   =   0b0000000000000000000001 #        1        0x1
REV_ARMv4T  =   0b0000000000000000000010 #        2        0x2
REV_ARMv5   =   0b0000000000000000000100 #        4        0x4
REV_ARMv5T  =   0b0000000000000000001000 #        8        0x8
REV_ARMv5E  =   0b0000000000000000010000 #       16        0x10
REV_ARMv5J  =   0b0000000000000000100000 #       32        0x20
REV_ARMv5TE =   0b0000000000000001000000 #       64        0x40
REV_ARMv6   =   0b0000000000000010000000 #      128        0x80
REV_ARMv6M  =   0b0000000000000100000000 #      256        0x100
REV_ARMv6T2 =   0b0000000000001000000000 #      512        0x200
REV_ARMv7A  =   0b0000000000010000000000 #     1024        0x400
REV_ARMv7R  =   0b0000000000100000000000 #     2048        0x800
REV_ARMv7M  =   0b0000000001000000000000 #     4096        0x1000
REV_ARMv7EM =   0b0000000010000000000000 #     8192        0x2000
REV_ARMv8A  =   0b0000000100000000000000 #    16384        0x4000
REV_ARMv8R  =   0b0000001000000000000000 #    32768        0x8000
REV_ARMv8M  =   0b0000010000000000000000 #    65536        0x10000

REV_ALL_ARMv4  = (REV_ARMv4 | REV_ARMv4T)
REV_ALL_ARMv5  = (REV_ARMv5 | REV_ARMv5T | REV_ARMv5E | REV_ARMv5J | REV_ARMv5TE)
REV_ALL_ARMv6  = (REV_ARMv6 | REV_ARMv6T2 | REV_ARMv6M)
REV_ALL_ARMv7  = (REV_ARMv7A | REV_ARMv7R | REV_ARMv7M | REV_ARMv7EM) 
REV_ALL_ARMv8  = (REV_ARMv8A | REV_ARMv8R | REV_ARMv8M)

#Todo - For easy instruction filtering add more like:
REV_ALL_TO_ARMv6 = REV_ALL_ARMv4 | REV_ALL_ARMv5 | REV_ALL_ARMv6
REV_ALL_FROM_ARMv6 = REV_ALL_ARMv6 | REV_ALL_ARMv7 | REV_ALL_ARMv8
#Note: since arm must be backwards compatible any depreciated commands
#Would be noted but not removed with the REV_ALL_TO_* if this is even used.
#These are put here for suggestion and comment for now

REV_ALL_ARM = (REV_ALL_ARMv4 | REV_ALL_ARMv5 | REV_ALL_ARMv6 | REV_ALL_ARMv7 | REV_ALL_ARMv8)

#Will be set below, THUMB16 up through v6 except v6T2, THUMB2 from v6T2 up, THUMBEE from v7 up.
REV_THUMB16 = REV_THUMB2  = REV_THUMBEE = 0   

ARCH_REVS = {}
#Iterate through all REV_ARM values and setup related combo values 
for name, val in list(globals().items()):
    if (not name.startswith('REV_ARM')):
        continue
    shortName = name[4:]
    #add to lookup dictionary
    ARCH_REVS[shortName] = val
    #setup thumb versions to Architecture versions
    if (int(shortName[4]) > 6 or shortName == 'ARMv6T2'):
        REV_THUMB2 = REV_THUMB2 | val
        if int(shortName[4])!= 6:
            REV_THUMBEE = REV_THUMBEE | val
    else:
        REV_THUMB16 = REV_THUMB16 | val
#Added thumbs to dictionary
ARCH_REVS['thumb16'] = REV_THUMB16
ARCH_REVS['thumb'] = REV_THUMB2
ARCH_REVS['thumbee'] = REV_THUMBEE
ARCH_REVSLEN = len(ARCH_REVS)

# The supported types of operand shifts (by the 2 bit field)
S_LSL = 0
S_LSR = 1
S_ASR = 2
S_ROR = 3
S_RRX = 4 # FIXME HACK XXX add this

opShifts = ('lsl', 'lsr', 'asr', 'ror', 'rrx')
movSuffixes = ('n', '', 'z', 'k')

extrepr = (
    'UXTB',
    'UXTH',
    'UXTW',
    'UXTX',
    'SXTB',
    'SXTH',
    'SXTW',
    'SXTX',
    'LSL',
)

EXT_UXTB    = 0
EXT_UXTH    = 1
EXT_UXTW    = 2
EXT_UXTX    = 3
EXT_SXTB    = 4
EXT_SXTH    = 5
EXT_SXTW    = 6
EXT_SXTX    = 7
EXT_LSL     = 8

ENDIAN_LSB = 0
ENDIAN_MSB = 1

iencs = (
    'IENC_DATA_SIMD',
    'IENC_LS_EXCL',
    'IENC_LS_NAPAIR_OFFSET',
    'IENC_LS_REGPAIR_POSTI',
    'IENC_LS_REGPAIR_OFFSET',
    'IENC_LS_REGPAIR_PREI',
    'IENC_LOG_SHFT_REG',
    'IENC_ADDSUB_SHFT_REG',
    'IENC_ADDSUB_EXT_REG',
    'IENC_SIMD_LS_MULTISTRUCT',
    'IENC_SIMD_LS_MULTISTRUCT_POSTI',
    'IENC_SIMD_LS_ONESTRUCT',
    'IENC_SIMD_LS_ONESTRUCT_POSTI',
    'IENC_PC_ADDR',
    'IENC_ADDSUB_IMM',
    'IENC_LOG_IMM',
    'IENC_MOV_WIDE_IMM',
    'IENC_BITFIELD',
    'IENC_EXTRACT',
    'IENC_CMP_BRANCH_IMM',
    'IENC_BRANCH_UNCOND_IMM',
    'IENC_BRANCH_COND_IMM',
    'IENC_EXCP_GEN',
    'IENC_SYS',
    'IENC_SME',
    'IENC_TEST_BRANCH_IMM',
    'IENC_BRANCH_UNCOND_REG',
    'IENC_LOAD_REG_LIT',
    'IENC_LS_REG_US_IMM',
    'IENC_LS_REG_UNSC_IMM',
    'IENC_LS_REG_IMM_POSTI',
    'IENC_LS_REG_UNPRIV',
    'IENC_LS_REG_IMM_PREI',
    'IENC_LS_REG_OFFSET',
    'IENC_LS_ATOMIC_MEM',
    'IENC_ADDSUB_CARRY',
    'IENC_COND_CMP_REG',
    'IENC_COND_CMP_IMM',
    'IENC_COND_SEL',
    'IENC_DATA_PROC_3',
    'IENC_DATA_PROC_2',
    'IENC_DATA_PROC_1',
    'IENC_FP_FP_CONV',
    'IENC_FP_COND_COMPARE',
    'IENC_FP_DP2',
    'IENC_FP_COND_SELECT',
    'IENC_FP_IMMEDIATE',
    'IENC_FP_COMPARE',
    'IENC_FP_DP1',
    'IENC_FP_INT_CONV',
    'IENC_FP_DP3',
    'IENC_SIMD_THREE_SAME',
    'IENC_SIMD_THREE_DIFF',
    'IENC_SIMD_TWOREG_MISC',
    'IENC_SIMD_ACROSS_LANES',
    'IENC_SIMD_COPY',
    'IENC_SIMD_VECTOR_IE',
    'IENC_SIMD_MOD_SHIFT_IMM',
    'IENC_SIMD_TBL_TBX',
    'IENC_SIMD_ZIP_UZP_TRN',
    'IENC_SIMD_EXT',
    'IENC_SIMD_SCALAR_THREE_SAME',
    'IENC_SIMD_SCALAR_THREE_DIFF',
    'IENC_SIMD_SCALAR_TWOREG_MISC',
    'IENC_SIMD_SCALAR_PAIRWISE',
    'IENC_SIMD_SCALAR_COPY',
    'IENC_SIMD_SCALAR_IE',
    'IENC_SIMD_SCALAR_SHIFT_IMM',
    'IENC_CRPYTO_AES',
    'IENC_CRYPTO_THREE_SHA',
    'IENC_CRYPTO_TWO_SHA',
    'IENC_SME_OUTER_PRODUCT_64_BIT',
    'IENC_SME_FP_OUTER_PRODUCT_32_BIT',
    'IENC_SME2_BINARY_OUTER_PRODUCT_32_BIT',
    'IENC_SME_INTEGER_OUTER_PRODUCT_32_BIT',
    'IENC_SME2_MULTI_VECTOR_MEMORY_CONTIGUOUS',
    'IENC_SME2_MULTI_VECTOR_MEMORY_STRIDED',
    'IENC_SME_MOVE_INTO_ARRAY',
    'IENC_SME_MOVE_FROM_ARRAY',
    'IENC_SME_ADD_VECTOR_TO_ARRAY',
    'IENC_SME_ZERO',
    'IENC_SME2_ZERO_LOOKUP_TABLE',
    'IENC_SME2_MOVE_LOOKUP_TABLE',
    'IENC_SME2_E0PAND_LOOKUP_TABLE_CONTIGUOUS',
    'IENC_SME2_MULTI_VECTOR_INDE0ED_ONE_REGISTER',
    'IENC_SME2_MULTI_VECTOR_INDE0ED_TWO_REGISTERS',
    'IENC_SME2_MULTI_VECTOR_INDE0ED_FOUR_REGISTERS',
    'IENC_SME2_MULTI_VECTOR_SVE_SELECT',
    'IENC_SME2_MULTI_VECTOR_SVE_CONSTRUCTIVE_BINARY',
    'IENC_SME2_MULTI_VECTOR_SVE_CONSTRUCTIVE_UNARY',
    'IENC_SME2_MULTI_VECTOR_MULTIPLE_VECTORS_SVE_DESTRUCTIVE_TWO_REGISTERS',
    'IENC_SME2_MULTI_VECTOR_MULTIPLE_VECTORS_SVE_DESTRUCTIVE_FOUR_REGISTERS',
    'IENC_SME2_MULTI_VECTOR_MULTIPLE_AND_SINGLE_SVE_DESTRUCTIVE_TWO_REGISTERS',
    'IENC_SME2_MULTI_VECTOR_MULTIPLE_AND_SINGLE_SVE_DESTRUCTIVE_FOUR_REGISTERS',
    'IENC_SME2_MULTI_VECTOR_MULTIPLE_AND_SINGLE_ARRAY_VECTORS',
    'IENC_SME2_MULTI_VECTOR_MULTIPLE_ARRAY_VECTORS_TWO_REGISTERS',
    'IENC_SME2_MULTI_VECTOR_MULTIPLE_ARRAY_VECTORS_FOUR_REGISTERS',
    'IENC_SME_MEMORY',
    'IENC_SVE_INT_MULTADD_PRED',
    'IENC_SVE_INT_BINARITH_PRED',
    'IENC_SVE_INT_REDUCTION',
    'IENC_SVE_BIT_SHIFT_PRED',
    'IENC_SVE_INT_ADDSUB_VEC_UNPRED',
    'IENC_SVE_BIT_LOG_UNPRED',
    'IENC_SVE_INDEX_GEN',
    'IENC_SVE_STACK_ALLOC',
    'IENC_SVE2_INT_MULT_UNPRED',
    'IENC_SVE_BIT_SHIFT_UNPRED',
    'IENC_SVE_ADDR_GEN',
    'IENC_SVE_INT_MISC_UNPRED',
    'IENC_SVE_ELEMENT_COUNT',
    'IENC_SVE_BIT_IMM',
    'IENC_SVE_INT_WIDE_IMM_UNPRED',
    'IENC_SVE_DUP_INDEXED',
    'IENC_SVE_UNALLOC',
    'IENC_SVE_TABLE_LOOKUP_THREE',
    'IENC_TBL_ENC',
    'IENC_SVE_PERMUTE_VEC_UNPRED',
    'IENC_SVE_PERMUTE_PRED',
    'IENC_SVE_PERMUTE_VEC_ELEMENTS',
    'IENC_SVE_PERMUTE_VEC_PRED',
    'IENC_SEL_VECTORS',
    'IENC_SVE_PERMUTE_VEC_EXTRACT',
    'IENC_SVE_PERMUTE_VEC_SEGMENTS',
    'IENC_SVE_INT_COMPARE_VEC',
    'IENC_SVE_INT_COMPARE_UNSIGNED_IMM',
    'IENC_SVE_INT_COMPARE_SIGNED_IMM',
    'IENC_SVE_PRED_LOGICAL_OPS',
    'IENC_SVE_PROP_BREAK',
    'IENC_SVE_PART_BREAK',
    'IENC_SVE_PRED_MISC',
    'IENC_SVE_INT_COMPARE_SCALARS',
    'IENC_SVE_BROAD_PRED_ELEMENT',
    'IENC_SVE_SCALAR_INT_COMPARE_PRED_CTR',
    'IENC_SVE_PRED_COUNT',
    'IENC_SVE_INC_DEC_PRED_CNT',
    'IENC_SVE_WRITE_FFR',
    'IENC_SVE_INT_MULTADD_UNPRED',
    'IENC_SVE2_INT_UNPRED',
    'IENC_SVE_INT_CLAMP',
    'IENC_SVE_MULT_INDEXED',
    'IENC_SVE_INT_TWO_WAY_DOT_PROD',
    'IENC_SVE_INT_TWO_WAY_DOT_PROD_INDEXED',
    'IENC_SVE2_WIDEN_INT_ARITH',
    'IENC_SVE_MISC',
    'IENC_SVE2_ACCUM',
    'IENC_SVE2_NARROWING',
    'IENC_SVE2_CHAR_MATCH',
    'IENC_SVE2_HIST_COMP_SEG',
    'IENC_HISTCNT',
    'IENC_SVE2_CRYPTO_EXT',
    'IENC_FCMLA_VEC',
    'IENC_FCADD',
    'IENC_SVE_FLOAT_CVT_PREC_ODD_ELEMENTS',
    'IENC_SVE2_FLOAT_PAIR_OPS',
    'IENC_SVE_FLOAT_MULTADD_INDEXED',
    'IENC_SVE_FLOAT_COMPLEX_MULTADD_INDEXED',
    'IENC_SVE_FLOAT_MULT_INDEXED',
    'IENC_FCLAMP',
    'IENC_SVE_FLOAT_WIDENING_MULTADD_INDEXED',
    'IENC_SVE_FLOAT_WIDENING_MULTADD',
    'IENC_SVE_FLOAT_MATRIX_MULT_ACC',
    'IENC_SVE_FLOAT_COMPARE_VECTORS',
    'IENC_SVE_FLOAT_ARITH_UNPRED',
    'IENC_SVE_FLOAT_ARITH_PRED',
    'IENC_SVE_FLOAT_UNARY_PRED',
    'IENC_SVE_FLOAT_RECURSE_REDUCTION',
    'IENC_SVE_FLOAT_UNARY_UNPRED',
    'IENC_SVE_FLOAT_COMPARE_ZERO',
    'IENC_SVE_FLOAT_ACC_REDUCTION',
    'IENC_SVE_FLOAT_MULTADD',
    'IENC_SVE_MEM_32GATHER_UNSIZED_CONTIG',
    'IENC_SVE_MEM_CONTIG_LOAD',
    'IENC_SVE_MEM_64GATHER',
    'IENC_SVE_MEM_CONTIG_STORE_UNSIZED_CONTIG',
    'IENC_SVE_MEM_NONTEMP_QUAD_SCATTER_STORE',
    'IENC_SVE_MEM_NONTEMP_MULTIREG_CONTIG_STORE',
    'IENC_SVE_MEM_SCATTER_OPT_SIGN_EXT',
    'IENC_SVE_MEM_SCATTER',
    'IENC_SVE_MEM_CONTIG_STORE_IMM_OFF',
    'IENC_RESERVED',
    'IENC_UNDEF',
)

IENC_MAX = len(iencs)

for ieidx, enc in enumerate(iencs):
    globals()[enc] = ieidx

instrnames = [
    'ADR',
    'ADRP',
    'ADD',
    'ADDS',
    'SUB',
    'SUBS',
    'CMP',
    'CMN',
    'AND',
    'ORR',
    'EOR',
    'ANDS',
    'MOV',
    'BFM',
    'EXTR',
    'B',
    'BL',
    'ST',
    'LD',
    'SVC',
    'HVC',
    'SMC',
    'BRK',
    'HLT',
    'DCPS1',
    'DCPS2',
    'DCPS3',
    'MSR',
    'HINT',
    'CSDB',
    'PSB',
    'CLREX',
    'DSB',
    'DMB',
    'ISB',
    'SYS',
    'SB',
    'MRS',
    'BR',
    'BLR',
    'RET',
    'ERET',
    'DRPS',
    'STXRB',
    'STLXRB',
    'LDXRB',
    'LDAXRB',
    'STLRB',
    'LDARB',
    'STXRH',
    'STLXRH',
    'LDXRH',
    'LDAXRH',
    'STLRH',
    'LDARH',
    'STXR',
    'STLXR',
    'STXP',
    'STLXP',
    'LDXR',
    'LDAXR',
    'LDXP',
    'LDAXP',
    'STLR',
    'LDAR',
    'LDR',
    'LDRSW',
    'PRFM',
    'STNP',
    'LDNP',
    'STGP',
    'STP',
    'LDP',
    'LDPSW',
    'LDUR',
    'STUR',
    'PRFUM',
    'STTR',
    'LDTR',
    'STR',
    'LDR',
    'ST4',
    'ST1',
    'LD4',
    'LD1',
    'ST3',
    'LD3',
    'ST2',
    'LD2',
    'BIC',
    'EO',
    'OR',
    'ADC',
    'SBC',
    'CCM',
    'CSEL',
    'CSINC',
    'CINC',
    'CSINV',
    'CINV',
    'CSET',
    'CSETM',
    'CSNEG',
    'CNEG',
    'MADD',
    'MSUB',
    'MNEG',
    'SMADDL',
    'SMULL',
    'SMSUBL',
    'SMNEGL',
    'UMADDL',
    'UMULL',
    'UMSUBL',
    'SMULH',
    'UMULH',
    'UDIV',
    'SDIV',
    'LSLV',
    'LSRV',
    'ASRV',
    'RORV',
    'ASR',
    'LSL',
    'LSR',
    'BFI',
    'BFXIL',
    'SBFX',
    'UBFX',
    'SBFIZ',
    'UBFIZ',
    'XTB',
    'XTH',
    'XTW',
    'SXTB',
    'UXTB',
    'SXTH',
    'UXTH',
    'SXTW',
    'UXTW',
    'LDADD',
    'LDADDA',
    'LDADDAL',
    'LDADDL',
    'LDADDB',
    'LDADDAB',
    'LDADDALB',
    'LDADDLB',
    'LDADDH',
    'LDADDAH',
    'LDADDALH',
    'LDADDLH',
    'LDCLR',
    'LDCLRA',
    'LDCLRAL',
    'LDCLRL',
    'LDCLRB',
    'LDCLRAB',
    'LDCLRALB',
    'LDCLRLB',
    'LDCLRH',
    'LDCLRAH',
    'LDCLRALH',
    'LDCLRLH',
    'LDEOR',
    'LDEORA',
    'LDEORAL',
    'LDEORL',
    'LDEORB',
    'LDEORAB',
    'LDEORALB',
    'LDEORLB',
    'LDEORH',
    'LDEORAH',
    'LDEORALH',
    'LDEORLH',
    'LDSET',
    'LDSETA',
    'LDSETAL',
    'LDSETL',
    'LDSETB',
    'LDSETAB',
    'LDSETALB',
    'LDSETLB',
    'LDSETH',
    'LDSETAH',
    'LDSETALH',
    'LDSETLH',
    'LDSMAX',
    'LDSMAXA',
    'LDSMAXAL',
    'LDSMAXL',
    'LDSMAXB',
    'LDSMAXAB',
    'LDSMAXALB',
    'LDSMAXLB',
    'LDSMAXH',
    'LDSMAXAH',
    'LDSMAXALH',
    'LDSMAXLH',
    'LDSMIN',
    'LDSMINA',
    'LDSMINAL',
    'LDSMINL',
    'LDSMINB',
    'LDSMINAB',
    'LDSMINALB',
    'LDSMINLB',
    'LDSMINH',
    'LDSMINAH',
    'LDSMINALH',
    'LDSMINLH',
    'LDUMAX',
    'LDUMAXA',
    'LDUMAXAL',
    'LDUMAXL',
    'LDUMAXB',
    'LDUMAXAB',
    'LDUMAXALB',
    'LDUMAXLB',
    'LDUMAXH',
    'LDUMAXAH',
    'LDUMAXALH',
    'LDUMAXLH',
    'LDUMIN',
    'LDUMINA',
    'LDUMINAL',
    'LDUMINL',
    'LDUMINB',
    'LDUMINAB',
    'LDUMINALB',
    'LDUMINLB',
    'LDUMINH',
    'LDUMINAH',
    'LDUMINALH',
    'LDUMINLH',
    'STADD',
    'STADDL',
    'STADDB',
    'STADDLB',
    'STADDH',
    'STADDLH',
    'STCLR',
    'STCLRL',
    'STCLRB',
    'STCLRLB',
    'STCLRH',
    'STCLRLH',
    'STEOR',
    'STEORL',
    'STEORB',
    'STEORLB',
    'STEORH',
    'STEORLH',
    'STSET',
    'STSETL',
    'STSETB',
    'STSETLB',
    'STSETH',
    'STSETLH',
    'STSMAX',
    'STSMAXL',
    'STSMAXB',
    'STSMAXLB',
    'STSMAXH',
    'STSMAXLH',
    'STSMIN',
    'STSMINL',
    'STSMINB',
    'STSMINLB',
    'STSMINH',
    'STSMINLH',
    'STUMAX',
    'STUMAXL',
    'STUMAXB',
    'STUMAXLB',
    'STUMAXH',
    'STUMAXLH',
    'LD64B',
    'ST64B',
    'ST64BV',
    'ST64BV0',
    'ST64B',
    'STUMIN',
    'STUMINL',
    'STUMINB',
    'STUMINLB',
    'STUMINH',
    'STUMINLH',
    'SWP',
    'SWPL',
    'SWPA',
    'SWPAL',
    'SWPB',
    'SWPAB',
    'SWPALB',
    'SWPLB',
    'SWPH',
    'SWPLH',
    'SWPAH',
    'SWPALH',
    'SWPLH',
    'CAS',
    'CASL',
    'CASA',
    'CASAL',
    'CASB',
    'CASLB',
    'CASAB',
    'CASALB',
    'CASH',
    'CASAH',
    'CASALH',
    'CASLH',
    'CASP',
    'CASPA',
    'CASPAL',
    'CASPL',
    'CRC32',
    'RBIT',
    'REV',
    'CL',
    'SCVTF',
    'UCVTF',
    'FCVTZS',
    'FCVTZU',
    'FCCMP',
    'FMUL',
    'FDIV',
    'FADD',
    'FSUB',
    'FMAX',
    'FMIN',
    'FMAXNM',
    'FMINNM',
    'FNMUL',
    'FCSEL',
    'FMADD',
    'FMSUB',
    'FNMADD',
    'FNMSUB',
    'FCMP',
    'FMOV',
    'FABS',
    'FNEG',
    'FSQRT',
    'FCVT',
    'FRINT',
    'CM',
    'SHL',
    'MUL',
    'AC',
    'ABD',
    'RECP',
    'SQRT',
    'ML',
    'ABS',
    'NEG',
    'NEGS',
    'XT',
    'CVT',
    'MAX',
    'MIN',
    'DUP',
    'MLA',
    'MLS',
    'SHR',
    'SRA',
    'SLI',
    'SHA1C',
    'SHA1P',
    'SHA1M',
    'SHA1SU0',
    'SHA256H',
    'SHA256H2',
    'SHA256SU1',
    'SHA1H',
    'SHA1SU1',
    'SHA256SU0',
    'AESE',
    'AESD',
    'AESMC',
    'AESIMC',
    'ADDHN',
    'SUBHN',
    'MUL',
    'ABA',
    'ABD',
    'UZP1',
    'TRN1',
    'ZIP1',
    'UZP2',
    'TRN2',
    'ZIP2',
    'CBZ',
    'CBNZ',
    'TBZ',
    'TBNZ',
    'AT',
    'DC',
    'IC',
    'TLBI',
    'BCC',
    'UDF',
    'NOP',
    'YIELD',
    'WFE',
    'WFI',
    'SEV',
    'SEVL',
    'DGH',
    'TST',
    'XPACI',
    'XPACD',
    'XPACLRI',
    'FMINNMP',
    'UQSHL',
]

ins_index = 85
for instr in instrnames:
    globals()['INS_' + instr] = ins_index
    ins_index += 1

#IFLAGS
IF_LE = 1 << 56
IF_HS = 1 << 55
IF_GE = 1 << 54
IF_HI = 1 << 53
IF_EQ = 1 << 52
IF_LT = 1 << 51
IF_GT = 1 << 50
IF_2 = 1 << 49
IF_C = 1 << 48
IF_32 = 1 << 44
IF_16 = 1 << 43
IF_P = 1 << 42
IF_PSR_S = 1 << 41
IF_N = 1 << 40
IF_Z = 1 << 39
IF_K = 1 << 38
IF_L = 1 << 37
IF_A = 1 << 36
IF_X = 1 << 35
IF_R = 1 << 34
IF_P = 1 << 33
IF_B = 1 << 32
IF_H = 1 << 15
IF_SW = 1 << 14
IF_S = 1 << 13
IF_W = 1 << 12
IF_E = 1 << 11
IF_M = 1 << 10
IF_I = 1 << 9
IF_U = 1 << 8

IFP_UQ = 1 << 62
IFP_U = 1 << 61
IFP_S = 1 << 60
IFP_P = 1 << 59
IFP_QD = 1 << 58
IFP_R = 1 << 57

IFS = (
    None,
    '.8b',
    '.16b',
    '.4h',
    '.8h',
    '.2s',
    '.4s',
    '.1d',
    '.2d',
)

for ifx in range(1, len(IFS)):
    ifs = IFS[ifx]
    gblname = "IFS" + ifs.upper().replace('.', '_')
    globals()[gblname] = ifx

PM_usr = 0b10000
PM_fiq = 0b10001
PM_irq = 0b10010
PM_svc = 0b10011
PM_mon = 0b10110
PM_abt = 0b10111
PM_hyp = 0b11010
PM_und = 0b11011
PM_sys = 0b11111

# reg stuff stolen from regs.py to support proc_modes
# these are in context of reg_table, not reg_data.  
#  ie. these are indexes into the lookup table.
REG_OFFSET_USR = 17 * (PM_usr&0xf)
REG_OFFSET_FIQ = 17 * (PM_fiq&0xf)
REG_OFFSET_IRQ = 17 * (PM_irq&0xf)
REG_OFFSET_SVC = 17 * (PM_svc&0xf)
REG_OFFSET_MON = 17 * (PM_mon&0xf)
REG_OFFSET_ABT = 17 * (PM_abt&0xf)
REG_OFFSET_HYP = 17 * (PM_hyp&0xf)
REG_OFFSET_UND = 17 * (PM_und&0xf)
REG_OFFSET_SYS = 17 * (PM_sys&0xf)
#REG_OFFSET_CPSR = 17 * 16
REG_OFFSET_CPSR = 16                    # CPSR is available in every mode, and PM_usr and PM_sys don't have an SPSR.

REG_SPSR_usr = REG_OFFSET_USR + 17
REG_SPSR_fiq = REG_OFFSET_FIQ + 17
REG_SPSR_irq = REG_OFFSET_IRQ + 17
REG_SPSR_svc = REG_OFFSET_SVC + 17
REG_SPSR_mon = REG_OFFSET_MON + 17
REG_SPSR_abt = REG_OFFSET_ABT + 17
REG_SPSR_hyp = REG_OFFSET_HYP + 17
REG_SPSR_und = REG_OFFSET_UND + 17
REG_SPSR_sys = REG_OFFSET_SYS + 17

REG_PC = 0xf
REG_SP = 0xd
REG_BP = None
REG_CPSR = REG_OFFSET_CPSR
REG_FLAGS = REG_OFFSET_CPSR    #same location, backward-compat name

proc_modes = { # mode_name, short_name, description, offset, mode_reg_count, PSR_offset, privilege_level
    PM_usr: ("User Processor Mode", "usr", "Normal program execution mode", REG_OFFSET_USR, 15, REG_SPSR_usr, 0),
    PM_fiq: ("FIQ Processor Mode", "fiq", "Supports a high-speed data transfer or channel process", REG_OFFSET_FIQ, 8, REG_SPSR_fiq, 1),
    PM_irq: ("IRQ Processor Mode", "irq", "Used for general-purpose interrupt handling", REG_OFFSET_IRQ, 13, REG_SPSR_irq, 1),
    PM_svc: ("Supervisor Processor Mode", "svc", "A protected mode for the operating system", REG_OFFSET_SVC, 13, REG_SPSR_svc, 1),
    PM_mon: ("Monitor Processor Mode", "mon", "Secure Monitor Call exception", REG_OFFSET_MON, 13, REG_SPSR_mon, 1),
    PM_abt: ("Abort Processor Mode", "abt", "Implements virtual memory and/or memory protection", REG_OFFSET_ABT, 13, REG_SPSR_abt, 1),
    PM_hyp: ("Hyp Processor Mode", "hyp", "Hypervisor Mode", REG_OFFSET_HYP, 13, REG_SPSR_hyp, 2),
    PM_und: ("Undefined Processor Mode", "und", "Supports software emulation of hardware coprocessor", REG_OFFSET_UND, 13, REG_SPSR_und, 1),
    PM_sys: ("System Processor Mode", "sys", "Runs privileged operating system tasks (ARMv4 and above)", REG_OFFSET_SYS, 15, REG_SPSR_sys, 1),
}

VFP_QWORD_REG_COUNT = 32    # ARMv8-A docs

'''
from envi.archs.aarch64.disasm import A64PreFetchOper
#Supported PRFOP options
prfop = (
    A64PreFetchOper(PLD, L1, KEEP),
    A64PreFetchOper(PLD, L1, STRM),
    A64PreFetchOper(PLD, L2, KEEP),
    A64PreFetchOper(PLD, L2, STRM),
    A64PreFetchOper(PLD, L3, KEEP),
    A64PreFetchOper(PLD, L3, STRM),
    '#uimm5',
    '#uimm5',
    A64PreFetchOper(PLI, L1, KEEP),
    A64PreFetchOper(PLI, L1, STRM),
    A64PreFetchOper(PLI, L2, KEEP),
    A64PreFetchOper(PLI, L2, STRM),
    A64PreFetchOper(PLI, L3, KEEP),
    A64PreFetchOper(PLI, L3, STRM),
    '#uimm5',
    '#uimm5',
    A64PreFetchOper(PST, L1, KEEP),
    A64PreFetchOper(PST, L1, STRM),
    A64PreFetchOper(PST, L2, KEEP),
    A64PreFetchOper(PST, L2, STRM),
    A64PreFetchOper(PST, L3, KEEP),
    A64PreFetchOper(PST, L3, STRM),
    '#uimm5',
    '#uimm5',
    '#uimm5',
    '#uimm5',
    '#uimm5',
    '#uimm5',
    '#uimm5',
    '#uimm5',
    '#uimm5',
    '#uimm5',
)
'''

sys_op_table = (
    (0b00001111000000, 'at', INS_AT),
    (0b10001111000000, 'at', INS_AT),
    (0b11001111000000, 'at', INS_AT),
    (0b00001111000001, 'at', INS_AT),
    (0b10001111000001, 'at', INS_AT),
    (0b11001111000001, 'at', INS_AT),
    (0b00001111000010, 'at', INS_AT),
    (0b00001111000011, 'at', INS_AT),
    (0b10001111000100, 'at', INS_AT),
    (0b10001111000101, 'at', INS_AT),
    (0b10001111000110, 'at', INS_AT),
    (0b10001111000111, 'at', INS_AT),
    (0b01101110100001, 'dc', INS_DC),
    (0b00001110110001, 'dc', INS_DC),
    (0b00001110110010, 'dc', INS_DC),
    (0b01101111010001, 'dc', INS_DC),
    (0b00001111010010, 'dc', INS_DC),
    (0b01101111011001, 'dc', INS_DC),
    (0b01101111110001, 'dc', INS_DC),
    (0b00001111110010, 'dc', INS_DC),
    (0b00001110001000, 'ic', INS_IC),
    (0b00001110101000, 'ic', INS_IC),
    (0b01101110101001, 'ic', INS_IC),
    (0b10010000000001, 'tlbi', INS_TLBI),
    (0b10010000000101, 'tlbi', INS_TLBI),
    (0b00010000011000, 'tlbi', INS_TLBI),
    (0b10010000011000, 'tlbi', INS_TLBI),
    (0b11010000011000, 'tlbi', INS_TLBI),
    (0b00010000011001, 'tlbi', INS_TLBI),
    (0b10010000011001, 'tlbi', INS_TLBI),
    (0b11010000011001, 'tlbi', INS_TLBI),
    (0b00010000011010, 'tlbi', INS_TLBI),
    (0b00010000011011, 'tlbi', INS_TLBI),
    (0b10010000011100, 'tlbi', INS_TLBI),
    (0b00010000011101, 'tlbi', INS_TLBI),
    (0b10010000011101, 'tlbi', INS_TLBI),
    (0b11010000011101, 'tlbi', INS_TLBI),
    (0b10010000011110, 'tlbi', INS_TLBI),
    (0b00010000011111, 'tlbi', INS_TLBI),
    (0b10010000100001, 'tlbi', INS_TLBI),
    (0b10010000100101, 'tlbi', INS_TLBI),
    (0b00010000111000, 'tlbi', INS_TLBI),
    (0b10010000111000, 'tlbi', INS_TLBI),
    (0b11010000111000, 'tlbi', INS_TLBI),
    (0b00010000111001, 'tlbi', INS_TLBI),
    (0b10010000111001, 'tlbi', INS_TLBI),
    (0b11010000111001, 'tlbi', INS_TLBI),
    (0b00010000111010, 'tlbi', INS_TLBI),
    (0b00010000111011, 'tlbi', INS_TLBI),
    (0b10010000111100, 'tlbi', INS_TLBI),
    (0b00010000111101, 'tlbi', INS_TLBI),
    (0b10010000111101, 'tlbi', INS_TLBI),
    (0b11010000111101, 'tlbi', INS_TLBI),
    (0b10010000111110, 'tlbi', INS_TLBI),
    (0b00010000111111, 'tlbi', INS_TLBI),
)

# AT instruction names, encoded as op1:CRm<0>:op2
at_op_table = {
    0b0000000: 'S1E1R',
    0b0000001: 'S1E1W',
    0b0000010: 'S1E0R',
    0b0000011: 'S1E0W',
    0b1000000: 'S1E2R',
    0b1000001: 'S1E2W',
    0b1000100: 'S12E1R',
    0b1000101: 'S12E1W',
    0b1000110: 'S12E0R',
    0b1000111: 'S12E0W',
    0b1100000: 'S1E3R',
    0b1100001: 'S1E3W',
    # Following require FEAT_PAN2:
    0b0001000: 'S1E1RP',
    0b0001001: 'S1E1WP'   
}

# DC instruction names, encoded as op1:crm:op2 
dc_op_table = {
    0b0000110001: 'IVAC',
    0b0000110010: 'ISW',
    0b0001010010: 'CSW',
    0b0001110010: 'CISW',
    0b0110100001: 'ZVA',
    0b0111010001: 'CVAC',
    0b0111011001: 'CVAU',
    0b0111110001: 'CIVAC',
    # Following require FEAT_MTE2:
    0b0000110011: 'IGVAC',
    0b0000110100: 'IGSW',
    0b0000110101: 'IGDVAC',
    0b0000110110: 'IGDSW',
    0b0001010100: 'CGSW',
    0b0001010110: 'CGDSW',
    0b0001110100: 'CIGSW',
    0b0001110110: 'CIGDSW ',
    # Following require FEAT_MTE:
    0b0110100011: 'GVA',
    0b0110100100: 'GZVA',
    0b0111010011: 'CGVAC',
    0b0111010101: 'CGDVAC',
    0b0111100011: 'CGVAP',
    0b0111100101: 'CGDVAP',
    0b0111101011: 'CGVADP',
    0b0111101101: 'CGDVADP',
    0b0111110011: 'CIGVAC',
    0b0111110101: 'CIGDVAC',
    # Following require FEAT_DPB:
    0b0111100001: 'CVAP',
    # Following require FEAT_DPB2:
    0b0111101001: 'CVADP',
    # Following require FEAT_MEC:
    0b1001110000: 'CIPAE',
    0b1001110111: 'CIGDPAE',
    # Following require FEAT_RME:
    0b1101110001: 'CIPAPA',
    0b1101110101: 'CIGDPAPA',
}

# IC instruction names, encoded as op1:crm:op2 
ic_op_table = {
    0b0000001000: 'IALLUIS',
    0b0000101000: 'IALLU',
    0b0110101001: 'IVAU'
}

# TLBI instruction names, encoded as op1:crn:crm:op2 
tlbi_op_table = {
    0b00010000011000: 'VMALLE1IS',
    0b00010000011001: 'VAE1IS',
    0b00010000011010: 'ASIDE1IS',
    0b00010000011011: 'VAAE1IS',
    0b00010000011101: 'VALE1IS',
    0b00010000011111: 'VAALE1IS',
    0b00010000111000: 'VMALLE1',
    0b00010000111001: 'VAE1',
    0b00010000011111: 'VAALE1IS',
    0b00010000111000: 'VMALLE1',
    0b00010000111001: 'VAE1',
    0b00010000111010: 'ASIDE1',
    0b00010000111011: 'VAAE1',
    0b00010000111101: 'VALE1',
    0b00010000111111: 'VAALE1',
    0b10010000000001: 'IPAS2E1IS',
    0b10010000000101: 'IPAS2LE1IS',
    0b10010000011000: 'ALLE2IS',
    0b10010000011001: 'VAE2IS',
    0b10010000011100: 'ALLE1IS',
    0b10010000011101: 'VALE2IS',
    0b10010000011110: 'VMALLS12E1IS',
    0b10010000100001: 'IPAS2E1',
    0b10010000100101: 'IPAS2LE1',
    0b10010000111000: 'ALLE2',
    0b10010000111001: 'VAE2',
    0b10010000111100: 'ALLE1',
    0b10010000111101: 'VALE2',
    0b10010000111110: 'VMALLS12E1',
    0b11010000011000: 'ALLE3IS',
    0b11010000011001: 'VAE3IS',
    0b11010000011101: 'VALE3IS',
    0b11010000111000: 'ALLE3',
    0b11010000111001: 'VAE3',
    0b11010000111101: 'VALE3',
    # Following require FEAT_TLBIOS:
    0b00010000001000: 'VMALLE1OS',
    0b00010000001001: 'VAE1OS',
    0b00010000001010: 'ASIDE1OS',
    0b00010000001011: 'VAAE1OS',
    0b00010000001101: 'VALE1OS',
    0b00010000001111: 'VAALE1OS',
    0b10010000001000: 'ALLE2OS',
    0b10010000001001: 'VAE2OS',
    0b10010000001100: 'ALLE1OS',
    0b10010000001100: 'VALE2OS',
    0b10010000011101: 'VMALLS12E1OS',
    0b10010000100000: 'IPAS2E1OS',
    0b10010000100100: 'IPAS2LE1OS',
    0b11010000001000: 'ALLE3OS',
    0b11010000001001: 'VAE3OS',
    0b11010000001101: 'VALE3OS',
    # Following requires FEAT_TLBIRANGE
    0b00010000010001: 'RVAE1IS',
    0b00010000010011: 'RVAAE1IS',
    0b00010000010101: 'RVALE1IS',
    0b00010000010111: 'RVAALE1IS',
    0b00010000101001: 'RVAE1OS',
    0b00010000101011: 'RVAAE1OS',
    0b00010000101101: 'RVALE1OS',
    0b00010000101111: 'RVAALE1OS',
    0b00010000110001: 'RVAE1',
    0b00010000110011: 'RVAAE1',
    0b00010000110101: 'RVALE1',
    0b00010000110111: 'RVAALE1',
    0b10010000000010: 'RIPAS2E1IS',
    0b10010000000110: 'RIPAS2LE1IS',
    0b10010000010001: 'RVAE2IS',
    0b10010000010101: 'RVALE2IS',
    0b10010000100010: 'RIPAS2E1',
    0b10010000100011: 'RIPAS2E1OS',
    0b10010000100110: 'RIPAS2LE1',
    0b10010000100111: 'RIPAS2LE1OS',
    0b10010000101001: 'RVAE20S',
    0b10010000101101: 'RVALE2OS',
    0b10010000110001: 'RVAE2',
    0b10010000110101: 'RVALE2',
    0b11010000010001: 'RVAE3IS',
    0b11010000010101: 'RVALE3IS',
    0b11010000101001: 'RVAE3OS',
    0b11010000101101: 'RVALE3OS',
    0b11010000110001: 'RAVE3',
    0b11010000110101: 'RVALE3',
    # Following require FEAT_XS:
    0b00010010001000: 'VMALLE1OSNXS',
    0b00010010001001: 'VAE1OSNXS',
    0b00010010001010: 'ASIDE1OSNXS',
    0b00010010001011: 'VAAESNXS',
    0b00010010001101: 'VALE1OSNXS',
    0b00010010001111: 'VAALE1OSNXS',
    0b00010010010001: 'RVAE1ISNXS',
    0b00010010010011: 'RVAAE1ISNXS',
    0b00010010010101: 'RVALE1ISNXS',
    0b00010010010111: 'RVAALE1ISNXS',
    0b00010010011000: 'VMALLE1ISNXS',
    0b00010010011001: 'VAE1ISNXS',
    0b00010010011010: 'ASIDE1ISNXS',
    0b00010010011011: 'VAAE1ISNXS',
    0b00010010011101: 'VALE1ISNXS',
    0b00010010011111: 'VAALE1ISNXS',
    0b00010010101001: 'RVAE1OSNXS',
    0b00010010101011: 'RVAAE1OSNXS',
    0b00010010101101: 'RVALE1OSNXS',
    0b00010010101111: 'RVAALE1OSNXS',
    0b00010010110001: 'RVAE1NXS',
    0b00010010110011: 'RVAAE1NXS',
    0b00010010110101: 'RVALE1NXS',
    0b00010010110111: 'RVAALE1NXS',
    0b00010010111000: 'VMALLE1NXS',
    0b00010010111001: 'VAE1NXS',
    0b00010010111010: 'ASIDE1NXS',
    0b00010010111011: 'VAAE1NXS',
    0b00010010111101: 'VALE1NXS',
    0b00010010111111: 'VAALE1NXS',
    0b10010010000001: 'IPAS2E1ISNXS',
    0b10010010000010: 'RIPAS2E1ISNXS',
    0b10010010000101: 'IPAS2LE1ISNXS',
    0b10010010000110: 'RIPAS2LE1ISNXS',
    0b10010010001000: 'ALLE2OSNXS',
    0b10010010001001: 'VAE2OSNXS',
    0b10010010001100: 'ALLE1OSNXS',
    0b10010010001101: 'VALE20SNXS',
    0b10010010001110: 'VMALLS12E1OSNXS',
    0b10010010010001: 'RVAE2ISNXS',
    0b10010010010101: 'RVALE2ISNXS',
    0b10010010011000: 'ALLE2ISNXS',
    0b10010010011001: 'VAE2ISNXS',
    0b10010010011100: 'ALLE1ISNXS',
    0b10010010011101: 'VALE2ISNXS',
    0b10010010011110: 'VMALLS12E1ISNXS',
    0b10010010100000: 'IPAS2E1OSNXS',
    0b10010010100001: 'IPAS2E1NXS',
    0b10010010100010: 'RIPAS2LE1NXS',
    0b10010010100011: 'RIPAS2E1OSNXS',
    0b10010010100100: 'IPAS2LE1OSNXS',
    0b10010010100101: 'IPAS2LE1NXS',
    0b10010010100110: 'RIPAS2LE1NXS',
    0b10010010100111: 'RIPAS2LE1OSNXS',
    0b10010010101001: 'RVAE2OSNXS',
    0b10010010101101: 'RVALE2OSNXS',
    0b10010010110001: 'RVAE2NXS',
    0b10010010110101: 'RAVLE2NXS',
    0b10010010111000: 'ALLE2NXS',
    0b10010010111001: 'VAE2NXS',
    0b10010010111100: 'ALLE1NXS',
    0b10010010111101: 'VALE2NXS',
    0b10010010111110: 'VMALLS12E1NXS',
    0b11010010001000: 'ALLE3OSNXS',
    0b11010010001001: 'VAE3OSNXS',
    0b11010010001101: 'VALE3OSNXS',
    0b11010010010001: 'RVAE3ISNXS',
    0b11010010010101: 'RVALE3ISNXS',
    0b11010010011000: 'ALLE3ISNXS',
    0b11010010011001: 'VAE3ISNXS',
    0b11010010011101: 'VALE3ISNXS',
    0b11010010101001: 'RVAE3OSNXS',
    0b11010010101101: 'RVALE3OSNXS',
    0b11010010110001: 'RVAE3NXS',
    0b11010010110101: 'RVALE3NXS',
    0b11010010111000: 'ALLE3NXS',
    0b11010010111001: 'VAE3NXS',
    0b11010010111101: 'VALE3NXS',
    # Following require FEAT_RME:
    0b11010000001100: 'PAALLOS',
    0b11010000100011: 'RPAOS',
    0b11010000100111: 'RPALOS',
    0b11010000111100: 'PAALL'
}

# Barrier operation assembler symbols, encoded in crm
bar_ops_table = {
    0b1111: 'sy',
    0b1110: 'st',
    0b1101: 'ld',
    0b1011: 'ish',
    0b1010: 'ishst',
    0b1001: 'ishld',
    0b0111: 'nsh',
    0b0110: 'nshst',
    0b0101: 'nshld',
    0b0011: 'osh',
    0b0010: 'oshst',
    0b0001: 'oshld'
}

psb_op = {
    0b0: 'csync'
}


sys_alias_op_tables = (
    at_op_table,
    dc_op_table,
    ic_op_table,
    tlbi_op_table,
    bar_ops_table,
    psb_op, 
)

barrier_option_table = (
    0,
    'OSHLD',
    'OSHST',
    'OSH',
    4,
    'NSHLD',
    'NSHST',
    'NSH',
    8,
    'ISHLD',
    'ISHST',
    'ISH',
    12,
    'LD',
    'ST',
    'SY',
)

b_cond_table = (
    ('b.eq', INS_BCC),
    ('b.ne', INS_BCC),
    ('b.hs', INS_BCC),
    ('b.lo', INS_BCC),
    ('b.mi', INS_BCC),
    ('b.pl', INS_BCC),
    ('b.vs', INS_BCC),
    ('b.vc', INS_BCC),
    ('b.hi', INS_BCC),
    ('b.ls', INS_BCC),
    ('b.ge', INS_BCC),
    ('b.lt', INS_BCC),
    ('b.gt', INS_BCC),
    ('b.le', INS_BCC),
    ('b.al', INS_BCC),
    ('b.nv^B', INS_BCC),
)

cond_table = (
    'EQ',
    'NE',
    'CS',
    'CC',
    'MI',
    'PL',
    'VS',
    'VC',
    'HI',
    'LS',
    'GE',
    'LT',
    'GT',
    'LE',
    'AL',
    'NV',
)

COND_EQ = 0
COND_NE = 1
COND_CS = 2
COND_CC = 3
COND_MI = 4
COND_PL = 5
COND_VS = 6
COND_VC = 7
COND_HI = 8
COND_LS = 9
COND_GE = 10
COND_LT = 11
COND_GT = 12
COND_LE = 13
COND_AL = 14
COND_NV = 15

inv_cond_table = (
        COND_NE,
        COND_EQ,
        COND_CC,
        COND_CS,
        COND_PL,
        COND_MI,
        COND_VC,
        COND_VS,
        COND_LS,
        COND_HI,
        COND_LT,
        COND_GE,
        COND_LE,
        COND_GT,
        COND_NV,
        COND_AL,
)


prefetchTypes = ["pld", "pli", "pst"]   
prefetchTargets = ["l1", "l2", "l3"]
prefetchPolicy = ["keep", "strm"]

# Pstatefield names for MSR using the "op1:op2:CRm" encoding
pstatefield_names = {
    0b0000110000: 'uao',
    0b0000110001: 'uao',
    0b0000110010: 'uao',
    0b0000110011: 'uao',
    0b0000110100: 'uao',
    0b0000110101: 'uao',
    0b0000110110: 'uao',
    0b0000110111: 'uao',
    0b0000111000: 'uao',
    0b0000111001: 'uao',
    0b0000111010: 'uao',
    0b0000111011: 'uao',
    0b0000111100: 'uao',
    0b0000111101: 'uao',
    0b0000111110: 'uao',
    0b0000111111: 'uao',
    0b0001000000: 'pan',
    0b0001000001: 'pan',
    0b0001000010: 'pan',
    0b0001000011: 'pan',
    0b0001000100: 'pan',
    0b0001000101: 'pan',
    0b0001000110: 'pan',
    0b0001000111: 'pan',
    0b0001001000: 'pan',
    0b0001001001: 'pan',
    0b0001001010: 'pan',
    0b0001001011: 'pan',
    0b0001001100: 'pan',
    0b0001001101: 'pan',
    0b0001001110: 'pan',
    0b0001001111: 'pan',
    0b0001010000: 'spsel',
    0b0001010001: 'spsel',
    0b0001010010: 'spsel',
    0b0001010011: 'spsel',
    0b0001010100: 'spsel',
    0b0001010101: 'spsel',
    0b0001010110: 'spsel',
    0b0001010111: 'spsel',
    0b0001011000: 'spsel',
    0b0001011001: 'spsel',
    0b0001011010: 'spsel',
    0b0001011011: 'spsel',
    0b0001011100: 'spsel',
    0b0001011101: 'spsel',
    0b0001011110: 'spsel',
    0b0001011111: 'spsel',
    0b0010000000: 'allint',
    0b0010000001: 'allint',
    0b0110100000: 'dit',
    0b0110100001: 'dit',
    0b0110100010: 'dit',
    0b0110100011: 'dit',
    0b0110100100: 'dit',
    0b0110100101: 'dit',
    0b0110100110: 'dit',
    0b0110100111: 'dit',
    0b0110101000: 'dit',
    0b0110101001: 'dit',
    0b0110101010: 'dit',
    0b0110101011: 'dit',
    0b0110101100: 'dit',
    0b0110101101: 'dit',
    0b0110101110: 'dit',
    0b0110101111: 'dit',
    0b0110110010: 'svcrsm',
    0b0110110011: 'svcrsm',
    0b0110110100: 'svcrza',
    0b0110110101: 'svcrza',
    0b0110110110: 'svcrsmza',
    0b0110110111: 'svcrsmza',
    0b0111000000: 'tco',
    0b0111000001: 'tco',
    0b0111000010: 'tco',
    0b0111000011: 'tco',
    0b0111000100: 'tco',
    0b0111000101: 'tco',
    0b0111000110: 'tco',
    0b0111000111: 'tco',
    0b0111001000: 'tco',
    0b0111001001: 'tco',
    0b0111001010: 'tco',
    0b0111001011: 'tco',
    0b0111001100: 'tco',
    0b0111001101: 'tco',
    0b0111001110: 'tco',
    0b0111001111: 'tco',
    0b0111100000: 'daifset',
    0b0111100001: 'daifset',
    0b0111100010: 'daifset',
    0b0111100011: 'daifset',
    0b0111100100: 'daifset',
    0b0111100101: 'daifset',
    0b0111100110: 'daifset',
    0b0111100111: 'daifset',
    0b0111101000: 'daifset',
    0b0111101001: 'daifset',
    0b0111101010: 'daifset',
    0b0111101011: 'daifset',
    0b0111101100: 'daifset',
    0b0111101101: 'daifset',
    0b0111101110: 'daifset',
    0b0111101111: 'daifset',
    0b0111110000: 'daifclr',
    0b0111110001: 'daifclr',
    0b0111110010: 'daifclr',
    0b0111110011: 'daifclr',
    0b0111110100: 'daifclr',
    0b0111110101: 'daifclr',
    0b0111110110: 'daifclr',
    0b0111110111: 'daifclr',
    0b0111111000: 'daifclr',
    0b0111111001: 'daifclr',
    0b0111111010: 'daifclr',
    0b0111111011: 'daifclr',
    0b0111111100: 'daifclr',
    0b0111111101: 'daifclr',
    0b0111111110: 'daifclr',
    0b0111111111: 'daifclr',
    0b0110010000: 'ssbs',
    0b0110010001: 'ssbs',
    0b0110010010: 'ssbs',
    0b0110010011: 'ssbs',
    0b0110010100: 'ssbs',
    0b0110010101: 'ssbs',
    0b0110010110: 'ssbs',
    0b0110010111: 'ssbs',
    0b0110011000: 'ssbs',
    0b0110011001: 'ssbs',
    0b0110011010: 'ssbs',
    0b0110011011: 'ssbs',
    0b0110011100: 'ssbs',
    0b0110011101: 'ssbs',
    0b0110011110: 'ssbs',
    0b0110011111: 'ssbs',
}

# Atomic operations mnem and opcode, encoded as size|v|a|r|o3|opc
ls_atomic_mem_table = {
    # Atomic add
    0b100000000: ['ldadd', INS_LDADD],
    0b100100000: ['ldadda', INS_LDADDA],
    0b100110000: ['ldaddal', INS_LDADDAL],
    0b100010000: ['ldaddl', INS_LDADDL],
    0b110000000: ['ldadd', INS_LDADD],
    0b110100000: ['ldadda', INS_LDADDA],
    0b110110000: ['ldaddal', INS_LDADDAL],
    0b110010000: ['ldaddl', INS_LDADDL],
    # Atomic add on byte
    0b000100000: ['ldaddab', INS_LDADDAB],
    0b000110000: ['ldaddalb', INS_LDADDALB],
    0b000000000: ['ldaddb', INS_LDADDB],
    0b000010000: ['ldaddlb', INS_LDADDLB],
    # Atomic add on halfword
    0b010100000: ['ldaddah', INS_LDADDAH],
    0b010110000: ['ldaddalh', INS_LDADDALH],
    0b010000000: ['ldaddh', INS_LDADDH],
    0b010010000: ['ldaddlh', INS_LDADDLH],
    # Atomic bit clear
    0b100000001: ['ldclr', INS_LDCLR],
    0b100100001: ['ldclra', INS_LDCLRA],
    0b100110001: ['ldclral', INS_LDCLRAL],
    0b100010001: ['ldclrl', INS_LDCLRL],
    0b110000001: ['ldclr', INS_LDCLR],
    0b110100001: ['ldclra', INS_LDCLRA],
    0b110110001: ['ldclral', INS_LDCLRAL],
    0b110010001: ['ldclrl', INS_LDCLRL],
    # Atomic bit clear on byte
    0b000100001: ['ldclrab', INS_LDCLRAB],
    0b000110001: ['ldclralb', INS_LDCLRALB],
    0b000000001: ['ldclrb', INS_LDCLRB],
    0b000010001: ['ldclrlb', INS_LDCLRLB],
    # Atomic bit clear on halfword
    0b010100001: ['ldclrah', INS_LDCLRAH],
    0b000110001: ['ldclralh', INS_LDCLRALH],
    0b000000001: ['ldclrh', INS_LDCLRH],
    0b000010001: ['ldclrlh', INS_LDCLRLH],
    # Atomic exclusive OR
    0b100000010: ['ldeor', INS_LDEOR],
    0b100100010: ['ldeora', INS_LDEORA],
    0b100110010: ['ldeoral', INS_LDEORAL],
    0b100010010: ['ldeorl', INS_LDEORL],
    0b110000010: ['ldeor', INS_LDEOR],
    0b110100010: ['ldeora', INS_LDEORA],
    0b110110010: ['ldeoral', INS_LDEORAL],
    0b110010010: ['ldeorl', INS_LDEORL],
    # Atomic exclusive OR on byte
    0b000100010: ['ldeorab', INS_LDEORAB],
    0b000110010: ['ldeoralb', INS_LDEORALB],
    0b000000010: ['ldeorb', INS_LDEORB],
    0b000010010: ['ldeorlb', INS_LDEORLB],
    # Atomic exclusive OR on halfword
    0b010100010: ['ldeorah', INS_LDEORAH],
    0b010110010: ['ldeoralh', INS_LDEORALH],
    0b010000010: ['ldeorh', INS_LDEORH],
    0b010010010: ['ldeorlh', INS_LDEORLH],
    # Atomic bit set
    0b100000011: ['ldset', INS_LDSET],
    0b100100011: ['ldseta', INS_LDSETA],
    0b100110011: ['ldsetal', INS_LDSETAL],
    0b100010011: ['ldsetl', INS_LDSETL],
    0b110000011: ['ldset', INS_LDSET],
    0b110100011: ['ldseta', INS_LDSETA],
    0b110110011: ['ldsetal', INS_LDSETAL],
    0b110010011: ['ldsetl', INS_LDSETL],
    # Atomic bit set on byte
    0b000100011: ['ldsetab', INS_LDSETAB],
    0b000110011: ['ldsetalb', INS_LDSETALB],
    0b000000011: ['ldsetb', INS_LDSETB],
    0b000010011: ['ldsetlb', INS_LDSETLB],
    # Atomic bit set on halfword
    0b010100011: ['ldsetah', INS_LDSETAH],
    0b010110011: ['ldsetalh', INS_LDSETALH],
    0b010000011: ['ldseth', INS_LDSETH],
    0b010010011: ['ldsetlh', INS_LDSETLH],
    # Atomic signed maximum
    0b100000100: ['ldsmax', INS_LDSMAX],
    0b100100100: ['ldsmaxa', INS_LDSMAXA],
    0b100110100: ['ldsmaxal', INS_LDSMAXAL],
    0b100010100: ['ldsmaxl', INS_LDSMAXL],
    0b110000100: ['ldsmax', INS_LDSMAX],
    0b110100100: ['ldsmaxa', INS_LDSMAXA],
    0b110110100: ['ldsmaxal', INS_LDSMAXAL],
    0b110010100: ['ldsmaxl', INS_LDSMAXL],
    # Atomic signed maximum on byte
    0b000000100: ['ldsmaxb', INS_LDSMAXB],
    0b000100100: ['ldsmaxab', INS_LDSMAXAB],
    0b000110100: ['ldsmaxalb', INS_LDSMAXALB],
    0b000010100: ['ldsmaxlb', INS_LDSMAXLB],
    # Atomic signed maximum on halfword
    0b010000100: ['ldsmaxh', INS_LDSMAXH],
    0b010100100: ['ldsmaxah', INS_LDSMAXAH],
    0b010110100: ['ldsmaxalh', INS_LDSMAXALH],
    0b010010100: ['ldsmaxlh', INS_LDSMAXLH],
    # Atomic signed minimum
    0b100000101: ['ldsmin', INS_LDSMIN],
    0b100100101: ['ldsmina', INS_LDSMINA],
    0b100110101: ['ldsminal', INS_LDSMINAL],
    0b100010101: ['ldsminl', INS_LDSMINL],
    0b110000101: ['ldsmin', INS_LDSMIN],
    0b110100101: ['ldsmina', INS_LDSMINA],
    0b110110101: ['ldsminal', INS_LDSMINAL],
    0b110010101: ['ldsminl', INS_LDSMINL],
    # Atomic signed minimum on byte
    0b000000101: ['ldsminb', INS_LDSMINB],
    0b000100101: ['ldsminab', INS_LDSMINAB],
    0b000110101: ['ldsminalb', INS_LDSMINALB],
    0b000010101: ['ldsminlb', INS_LDSMINLB],
    # Atomic signed minimum on halfword
    0b010000101: ['ldsminh', INS_LDSMINH],
    0b010100101: ['ldsminah', INS_LDSMINAH],
    0b010110101: ['ldsminalh', INS_LDSMINALH],
    0b010010101: ['ldsminlh', INS_LDSMINLH],
    # Atomic unsigned maximum
    0b100000110: ['ldumax', INS_LDUMAX],
    0b100100110: ['ldumaxa', INS_LDUMAXA],
    0b100110110: ['ldumaxal', INS_LDUMAXAL],
    0b100010110: ['ldumaxl', INS_LDUMAXL],
    0b110000110: ['ldumax', INS_LDUMAX],
    0b110100110: ['ldumaxa', INS_LDUMAXA],
    0b110110110: ['ldumaxal', INS_LDUMAXAL],
    0b110010110: ['ldumaxl', INS_LDUMAXL],
    # Atomic unsigned maximum on byte
    0b000000110: ['ldumaxb', INS_LDUMAXB],
    0b000100110: ['ldumaxab', INS_LDUMAXAB],
    0b000110110: ['ldumaxalb', INS_LDUMAXALB],
    0b000010110: ['ldumaxlb', INS_LDUMAXLB],
    # Atomic unsigned maximum on halfword
    0b010000110: ['ldumaxh', INS_LDUMAXH],
    0b010100110: ['ldumaxah', INS_LDUMAXAH],
    0b010110110: ['ldumaxalh', INS_LDUMAXALH],
    0b010010110: ['ldumaxlh', INS_LDUMAXLH],
    # Atomic unsigned minimum
    0b100000111: ['ldumin', INS_LDUMIN],
    0b100100111: ['ldumina', INS_LDUMINA],
    0b100110111: ['lduminal', INS_LDUMINAL],
    0b100010111: ['lduminl', INS_LDUMINL],
    0b110000111: ['ldumin', INS_LDUMIN],
    0b110100111: ['ldumina', INS_LDUMINA],
    0b110110111: ['lduminal', INS_LDUMINAL],
    0b110010111: ['lduminl', INS_LDUMINL],
    # Atomic unsigned minimum on byte
    0b000000111: ['lduminb', INS_LDUMINB],
    0b000100111: ['lduminab', INS_LDUMINAB],
    0b000110111: ['lduminalb', INS_LDUMINALB],
    0b000010111: ['lduminlb', INS_LDUMINLB],
    # Atomic unsigned minimum on halfword
    0b010000111: ['lduminh', INS_LDUMINH],
    0b010100111: ['lduminah', INS_LDUMINAH],
    0b010110111: ['lduminalh', INS_LDUMINALH],
    0b010010111: ['lduminlh', INS_LDUMINLH],
    # Single-copy Atomic 64-byte load
    0b110001101: ['ld64b', INS_LD64B],
    # Single-copy Atomic 64-byte store
    0b110001001: ['st64b', INS_ST64B],
    # Single-copy Atomic 64-byte Store with return
    0b110001011: ['st64bv', INS_ST64BV],
    # Single-copy Atomic 64-byte EL0 Store with Return 
    0b110001010: ['st64bv0', INS_ST64BV0],
    # Swap word or doubleword
    0b100001000: ['swp', INS_SWP],
    0b100101000: ['swpa', INS_SWPA],
    0b100111000: ['swpal', INS_SWPAL],
    0b100011000: ['swpl', INS_SWPL],
    0b110001000: ['swp', INS_SWP],
    0b110101000: ['swpa', INS_SWPA],
    0b110111000: ['swpal', INS_SWPAL],
    0b110011000: ['swpl', INS_SWPL],
    # Swap byte
    0b000001000: ['swpb', INS_SWPB],
    0b000101000: ['swpab', INS_SWPAB],
    0b000111000: ['swpalb', INS_SWPALB],
    0b000011000: ['swplb', INS_SWPLB],
    # Swap halfword
    0b010001000: ['swpb', INS_SWPH],
    0b010101000: ['swpab', INS_SWPAH],
    0b010111000: ['swpalb', INS_SWPALH],
    0b010011000: ['swplb', INS_SWPLH],
}

# Aliased Atomic operation mnem and opcode, encoded as size|v|a|r|o3|opc
ls_atomic_mem_alias_table = {
    # Atomic add, without return 
    0b100000000: ['stadd', INS_STADD],
    0b100010000: ['staddl', INS_STADDL],
    0b110000000: ['stadd', INS_STADD],
    0b110010000: ['staddl', INS_STADDL],
    # Atomic add on byte, without return
    0b000000000: ['staddb', INS_STADDB],
    0b000010000: ['staddlb', INS_STADDLB],
    # Atomic add on halfword, without return
    0b010000000: ['staddh', INS_STADDH],
    0b010010000: ['staddlh', INS_STADDLH],
    # Atomic bit clear, without return
    0b100000001: ['stclr', INS_STCLR],
    0b100010001: ['stclrl', INS_STCLRL],
    0b110000001: ['stclr', INS_STCLR],
    0b110010001: ['stclrl', INS_STCLRL],
    # Atomic bit clear on byte, without return
    0b000000001: ['stclrb', INS_STCLRB],
    0b000010001: ['stclrlb', INS_STCLRLB],
    # Atomic bit clear on halfword, without return
    0b010000001: ['stclrh', INS_STCLRH],
    0b010010001: ['stclrlh', INS_STCLRLH],
    # Atomic exclusive-OR, without return 
    0b100000010: ['steor', INS_STEOR],
    0b100010010: ['steorl', INS_STEORL],
    0b110000010: ['steor', INS_STEOR],
    0b110010010: ['steorl', INS_STEORL],
    # Atomic exclusive-OR on byte, without return 
    0b000000010: ['steorb', INS_STEORB],
    0b000010010: ['steorlb', INS_STEORLB],
    # Atomic exclusive-OR on halfword, without return
    0b010000010: ['steorh', INS_STEORH],
    0b010010010: ['steorlh', INS_STEORLH],
    # Atomic signed maximum, without return
    0b100000100: ['stsmax', INS_STSMAX],
    0b100010100: ['stsmaxl', INS_STSMAXL],
    0b110000100: ['stsmax', INS_STSMAX],
    0b110010100: ['stsmaxl', INS_STSMAXL],
    # Atomic signed maximum on byte, without return
    0b000000100: ['stsmaxb', INS_STSMAXB],
    0b000010100: ['stsmaxlb', INS_STSMAXLB],
    # Atomic signed maximum on halfword, without return
    0b010000100: ['stsmaxh', INS_STSMAXH],
    0b010010100: ['stsmaxlh', INS_STSMAXLH],
    # Atomic signed minimum, without return
    0b100000101: ['stsmin', INS_STSMIN],
    0b100010101: ['stsminl', INS_STSMINL],
    0b110000101: ['stsmin', INS_STSMIN],
    0b110010101: ['stsminl', INS_STSMINL],
    # Atomic signed minimum on byte, without return
    0b000000101: ['stsminb', INS_STSMINB],
    0b000010101: ['stsminlb', INS_STSMINLB],
    # Atomic signed minimum on halfword, without return
    0b010000101: ['stsminh', INS_STSMINH],
    0b010010101: ['stsminlh', INS_STSMINLH],
    # Atomic unsigned maximum, without return 
    0b100000110: ['stumax', INS_STUMAX],
    0b100010110: ['stumaxl', INS_STUMAXL],
    0b110000110: ['stumax', INS_STUMAX],
    0b110010110: ['stumaxl', INS_STUMAXL],
    # Atomic unsigned maximum on byte, without return 
    0b000000110: ['stumaxb', INS_STUMAXB],
    0b000010110: ['stumaxlb', INS_STUMAXLB],
    # Atomic unsigned maximum on halfword, without return 
    0b010000110: ['stumaxh', INS_STUMAXH],
    0b010010110: ['stumaxlh', INS_STUMAXLH],
    # Atomic unsigned minimum, without return
    0b100000111: ['stumin', INS_STUMIN],
    0b100010111: ['stuminl', INS_STUMINL],
    0b110000111: ['stumin', INS_STUMIN],
    0b110010111: ['stuminl', INS_STUMINL],
    # Atomic unsigned minimum on byte, without return
    0b000000111: ['stuminb', INS_STUMINB],
    0b000010111: ['stuminlb', INS_STUMINLB],
    # Atomic unsigned minimum on halfword, without return
    0b010000111: ['stuminh', INS_STUMINH],
    0b010010111: ['stuminlh', INS_STUMINLH],
}


# Compare and Swap operation mnem and opcode, encoded as size|v|a|r|o3|opc
ls_comp_swap_table = {
    # Compare and Swap word or doubleword
    0b100100111: ['cas', INS_CAS],
    0b100110111: ['casa', INS_CASA],
    0b100111111: ['casal', INS_CASAL],
    0b100101111: ['casl', INS_CASL],
    0b110100111: ['cas', INS_CAS],
    0b110110111: ['casa', INS_CASA],
    0b110111111: ['casal', INS_CASAL],
    0b110101111: ['casl', INS_CASL],
    # Compare and Swap byte
    0b000100111: ['casb', INS_CASB],
    0b000110111: ['casab', INS_CASAB],
    0b000111111: ['casalb', INS_CASALB],
    0b000101111: ['caslb', INS_CASLB],
    # Compare and Swap halfword
    0b010100111: ['cash', INS_CASH],
    0b010110111: ['casah', INS_CASAH],
    0b010111111: ['casalh', INS_CASALH],
    0b010101111: ['caslh', INS_CASLH],
    # Compare and Swap Pair of words or doublewords
    0b000000111: ['casp', INS_CASP],
    0b000010111: ['caspa', INS_CASPA],
    0b000011111: ['caspal', INS_CASPAL],
    0b000001111: ['caspl', INS_CASPL],
    0b010000111: ['casp', INS_CASP],
    0b010010111: ['caspa', INS_CASPA],
    0b010011111: ['caspal', INS_CASPAL],
    0b010001111: ['caspl', INS_CASPL],
}



import os
import logging
import binascii

LOG_FORMAT = '%(asctime)s:%(levelname)s:%(name)s: %(message)s' \
             '[%(filename)s:%(funcName)s:%(lineno)s]'

EMULOG = 9
SHITE = 8
MIRE = 7

LOG_LEVELS = (
    logging.CRITICAL,
    logging.ERROR,
    logging.WARN,
    logging.INFO,
    logging.DEBUG,
    EMULOG,
    SHITE,
    MIRE,
)

def getLogLevelFromString(loglvlstr):
    '''
    Translate a string to a Logging constant
    '''
    if hasattr(logging, loglvlstr):
        return getattr(logging, loglvlstr)
    return globals().get(loglvlstr)


def initLogging(logger, level=None, fmt=LOG_FORMAT):
    '''
    Setup logging and log levels
    '''
    if level is None:
        level = os.environ.get('VIV_LOGLEVEL')

    if type(level) == str:
        level = getLogLevelFromString(level)

    if level:
        if level not in LOG_LEVELS:
            raise ValueError('Invalid log level of %r' % level)
    else:
        level = 'ERROR'

    # Setup logging
    logging.addLevelName(EMULOG, 'EMULOG')
    logging.addLevelName(SHITE, 'SHITE')
    logging.addLevelName(MIRE, 'MIRE')
    logging.basicConfig(level=level, format=fmt)

    # Set log level for current logger
    logger.setLevel(level)


def hexify(byts):
    '''
    Return a series of bytes as a string of hex characters. The return type will be a python string.
    '''
    return binascii.hexlify(byts).decode('utf-8')




def spline4p(t, p_1, p0, p1, p2):
    x_1a = t * ((2-t) * t - 1) * p_1[0]
    x_1b = t * ((2-t) * t - 1) * p_1[1]

    x0a  = (t * t * (3 * t - 5) + 2) * p0[0]
    x0b  = (t * t * (3 * t - 5) + 2) * p0[1]

    x1a  = t * ((4 - 3 * t) * t + 1) * p1[0]
    x1b  = t * ((4 - 3 * t) * t + 1) * p1[1]

    x2a  = (t - 1) * t * t * p2[0]
    x2b  = (t - 1) * t * t * p2[1]

    final = ( (x_1a + x0a + x1a + x2a) / 2,
              (x_1b + x0b + x1b + x2b) / 2)

    return final

def calculate_catmullrom(points, steps = 30):
    ret = []
    percents = [ step / float(steps) for step in range(steps) ]
    for i in range(1, len(points) - 2):
        for t in percents:
            ret.append(spline4p(t, points[i-1], points[i], points[i+1], points[i+2]))
    return ret



import sys

import cobra.unittests

# Import order here is by test run order!
import cobra.unittests.basictest as c_basictest
import cobra.unittests.msgpacktest as c_msgpacktest
import cobra.unittests.reftest as c_reftest
import cobra.unittests.authtest as c_authtest
import cobra.unittests.shadowtest as c_shadowtest

def main():
    cobra.unittests.runUnitTests()

if __name__ == '__main__':
    sys.exit(main())



import vstruct

from vstruct.primitives import *
from vstruct.defs.macho.const import *

class nlist(vstruct.VStruct):
    '''
    A symbol table entry in a Mach-O binary is called an nlist.
    '''
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.n_strx     = v_uint32()    # index into the string table
        self.n_type     = v_uint8()     # type flag (see const...)
        self.n_sect     = v_uint8()     # section number or NO_SECT (index from 1...)
        self.n_desc     = v_uint16()    # desription (see const...)
        self.n_value    = v_uint32()    # value of this symbol (or stab offset)

class nlist64(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.n_strx     = v_uint32()
        self.n_type     = v_uint8()
        self.n_sect     = v_uint8()
        self.n_desc     = v_uint16()
        self.n_value    = v_uint64()




import os
import unittest

import PE

def getTestPath(*paths):
    '''
    Return the join'd path to a file in the vivtestfiles repo
    by using the environment variable "VIVTESTFILES"

    ( raises SkipTest if env var is not present )
    '''
    testdir = os.getenv('VIVTESTFILES')
    if not testdir:
        raise unittest.SkipTest('VIVTESTFILES env var not found!')
    testdir = os.path.abspath(testdir)
    return os.path.join(testdir, *paths)

class PEResourceTests(unittest.TestCase):

    def test_repr(self):
        path = getTestPath('windows','amd64','dbghelp.dll')
        pe = PE.peFromFileName(path)
        pe_repr = repr(pe)

        self.assertIn("DllName: 'dbghelp.dll'", pe_repr)
        self.assertIn("PDB Path: 'dbghelp.pdb'", pe_repr)
        self.assertIn("00000000 (08)   Name: .rsrc", pe_repr)
        self.assertIn("00000048 (02)     MajorSubsystemVersion: 0x00000005 (5)", pe_repr)





import os

from PyQt5 import QtCore
from PyQt5.QtWidgets import *

import envi.cli as e_cli
import envi.qt.memcanvas as e_q_memcanvas

import vqt.colors as vq_colors
import vqt.hotkeys as vq_hotkeys

from vqt.basics import *
from vqt.main import idlethread,workthread

class VQInput(vq_hotkeys.HotKeyMixin, QLineEdit):

    def __init__(self, parent=None):
        QLineEdit.__init__(self, parent=parent)
        vq_hotkeys.HotKeyMixin.__init__(self)

        self.history = []
        self.histidx = 0

        self.addHotKey('up','cli:history:prev')
        self.addHotKey('down','cli:history:next')

    def useHistory(self, delta):

        if delta < 0 and self.histidx == 0:
            return

        if delta > 0 and len(self.history) <= self.histidx+delta:
            self.histidx = len(self.history)
            self.clear()
            return

        self.histidx += delta
        htext = self.history[self.histidx]
        self.setText(htext)

    def addHistory(self, histcmd):
        if histcmd:
            self.history.append(histcmd)
            self.histidx = len(self.history)

    @vq_hotkeys.hotkey('cli:history:prev')
    def keyCodeUp(self,*args):
        self.useHistory(-1)

    @vq_hotkeys.hotkey('cli:history:next')
    def keyCodeDown(self):
        self.useHistory(1)

    def loadHistory(self, filename):
        if os.path.isfile(filename):
            with open(filename, 'r') as f:
                hist = f.readlines()[-1000:]
            self.history = [ x.strip() for x in hist ]
            self.histidx = len(self.history)

    def saveHistory(self, filename):
        histbuf = '\n'.join( self.history[-1000:] )
        with open(filename, 'w') as f:
            f.write(histbuf)


class VQCli(QWidget):
    '''
    A Qt class to wrap and emulate a Cmd object.
    '''

    sigCliQuit = QtCore.pyqtSignal()

    def __init__(self, cli, parent=None):
        QWidget.__init__(self, parent=parent)
        self.cli = cli

        self.input = VQInput(self)

        # Create our output window...
        self.output = QTextEdit(self)
        # If it's an EnviCli, let's over-ride the canvas right away.
        if isinstance(cli, e_cli.EnviCli):
            self.output.close()
            self.output = self.initMemoryCanvas(cli.memobj, syms=cli.symobj)
            self.output.setScrolledCanvas(True)
            cli.setCanvas(self.output)

        self.setStyleSheet(vq_colors.getDefaultColors())

        self.setLayout(self.getCliLayout())

        self.input.returnPressed.connect(self.returnPressedSlot)

        self.resize(250, 150)

    def initMemoryCanvas(self, memobj, syms=None):
        return e_q_memcanvas.VQMemoryCanvas(memobj, syms=syms, parent=self)

    def getCliLayout(self):
        return VBox(self.output, self.input)

    def returnPressedSlot(self):
        cmd = str(self.input.text())
        self.input.clear()
        self.input.addHistory(cmd)
        self.output.addText('> %s\n' % cmd)
        workthread(self.onecmd)(cmd)

    def onecmd(self, cmdline):
        if self.cli.onecmd(cmdline):
            self._emit_quit()

    @idlethread
    def _emit_quit(self):
        # A way to emit the "quit" signal from threads other than the
        # qt main thread.
        self.sigCliQuit.emit()



import logging
import unittest

import vivisect.cli as viv_cli
import vivisect.tests.helpers as helpers

logger = logging.getLogger(__name__)

path = ('raw', 'msp430', 'blink.srec')


class SrecTests(unittest.TestCase):

    def test_srec(self):
        fn = helpers.getTestPath(*path)
        vw = viv_cli.VivCli()
        vw.config.viv.parsers.srec.arch = 'msp430'
        vw.loadFromFile(fn)

        vw.makeFunction(0x4000)

        self.assertEqual(vw.getFunction(0x4000), 0x4000)
        self.assertEqual(vw.getFunction(0x4050), 0x4000)
        self.assertEqual(vw.getFunction(0x4060), 0x405e)
        self.assertEqual(vw.getFunction(0x4068), 0x405e)



import unittest
from io import BytesIO

import cobra


class NewObjectReturn:

    @cobra.newobj
    def open(self):
        return BytesIO('asdf'.encode('utf-8'))


class CobraNewObjTest(unittest.TestCase):

    def test_cobra_newobj(self):

        daemon = cobra.CobraDaemon(port=60500, msgpack=True)
        objname = daemon.shareObject(NewObjectReturn())
        daemon.fireThread()

        t = cobra.CobraProxy('cobra://localhost:60500/%s?msgpack=1' % objname)

        with t.open() as fd:
            self.assertEqual(fd.read(), b'asdf')

        self.assertEqual(len(daemon.shared.keys()), 1)
        daemon.stopServer()




"""
Cobra's built in clustering framework
"""

import gc
import sys
import time
import struct
import socket
import logging
import traceback
import threading
import subprocess
import multiprocessing
import urllib.request as url_req

import cobra
from . import dcode

logger = logging.getLogger(__name__)

queen_port = 32124

cluster_port = 32123
cluster_ip = "224.69.69.69"

sub_cmd = """
import cobra.cluster
cobra.cluster.getAndDoWork("%s", docode=%s)
"""

class InvalidInProgWorkId(Exception):
    def __init__(self, workid):
        Exception.__init__(self, "Work ID %d is not valid" % workid)
        self.workid = workid

class ClusterWork(object):
    """
    Extend this object to create your own work units.  Do it in
    a proper module (and not __main__ to be able to use this
    in conjunction with cobra.dcode).
    """
    def __init__(self, timeout=None):
        object.__init__(self)
        self.id = None # Set by adding to the server
        self.server = None # Set by ClusterClient
        self.starttime = 0
        self.endtime = 0 # Both are set by worker before and after work()
        self.timeout = timeout
        self.touchtime = None
        self.excinfo = None # Will be exception traceback on work unit fail.

    def touch(self): # heh...
        """
        Update the internal "touch time" which is used by the timeout
        subsystem to see if this work unit has gone too long without
        making progress...
        """
        self.touchtime = time.time()

    def isTimedOut(self):
        """
        Check if this work unit is timed out.
        """
        if self.timeout is None:
            return False
        if self.touchtime is None:
            return False
        if self.endtime != 0:
            return False
        return (self.touchtime + self.timeout) < time.time()

    def work(self):
        """
        Actually do the work associated with this work object.
        """
        for i in range(10):
            self.setCompletion(i*10)
            self.setStatus("Sleeping: %d" % i)
            time.sleep(1)

    def done(self):
        """
        This is called back on the server once a work unit
        is complete and returned.
        """
        pass

    def setCompletion(self, percent):
        """
        Work units may call this whenever they like to
        tell the server how far along their work they are.
        """
        self.touch()
        self.server.setWorkCompletion(self.id, percent)

    def setStatus(self, status):
        """
        Work units may call this to inform the server of
        their status.
        """
        self.touch()
        self.server.setWorkStatus(self.id, status)

    def openSharedFile(self, filename):
        '''
        A helper API to open a file like object on the server.

        Example:
            fd = self.openSharedFile('/foo/bar/baz')
            fbytes = fd.read()

        NOTE: The server must use shareFileToWorkers().
        '''
        uri = self.server.openSharedFile( filename )
        return cobra.CobraProxy(uri)

    def getServerObject(self, objname):
        '''
        Get a CobraProxy for an object shared by the ClusterServer (eg. a VivWorkspace)
        '''
        uri = self.server.getServerObjectUri(objname)
        proxy = cobra.CobraProxy(uri, timeout=60, retrymax=3)
        return proxy

class ClusterCallback:
    """
    Place one of these in the ClusterServer to get synchronous
    event information about what's going on in the cluster server.
    (mostly for the GUI).
    """

    def workAdded(self, server, work):
        pass
    def workGotten(self, server, work):
        pass
    def workStatus(self, server, workid, status):
        pass
    def workCompletion(self, server, workid, completion):
        pass
    def workDone(self, server, work):
        pass
    def workFailed(self, server, work):
        pass
    def workTimeout(self, server, work):
        pass
    def workCanceled(self, server, work):
        pass

class VerboseCallback(ClusterCallback):
    # This is mostly for testing...
    def workAdded(self, server, work):
        logger.debug("WORK ADDED: %d", work.id)

    def workGotten(self, server, work):
        logger.debug("WORK GOTTEN: %d", work.id)

    def workStatus(self, server, workid, status):
        logger.debug("WORK STATUS: (%d) %s", (workid, status))

    def workCompletion(self, server, workid, completion):
        logger.debug("WORK COMPLETION: (%d) %d%%", workid, completion)

    def workDone(self, server, work):
        logger.debug("WORK DONE: %d", work.id)

    def workFailed(self, server, work):
        logger.debug("WORK FAILED: %d", work.id)

    def workTimeout(self, server, work):
        logger.debug("WORK TIMEOUT: %d", work.id)

    def workCanceled(self, server, work):
        logger.debug("WORK CANCELED %d", work.id)

import collections

class ClusterServer:

    def __init__(self, name, maxsize=None, docode=False, bindsrc="", cobrad=None):
        """
        The cluster server is the core of the code that manages work units.

        Arguments:
            maxsize - How big should the work queue be before add blocks
            docode  - Should we also be a dcode server?
            bindsrc - Should we bind a src IP for our multicast announcements?
            cobrad  - Should we use an existing cobra daemon to share our objects?
        """
        self.go = True
        self.name = name
        self.queens = []
        self.nextwid = 0
        self.inprog = {}
        self.sharedfiles = {}
        self.maxsize = maxsize
        self.queue = collections.deque()
        self.qcond = threading.Condition()
        self.widiter = iter(range(999999999))

        # Initialize a cobra daemon if needed
        if cobrad is None:
            cobrad = cobra.CobraDaemon(host="", port=0)
        self.cobrad = cobrad
        self.cobraname = self.cobrad.shareObject(self)

        # Setup our transmission socket
        self.sendsock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sendsock.bind((bindsrc, 0))

        # Set this to a ClusterCallback extension if
        # you want notifications.
        self.callback = None

        if docode:
            dcode.enableDcodeServer(daemon=self.cobrad)

        # Fire the timeout monitor thread...
        thr = threading.Thread(target=self.timerThread, daemon=True)
        thr.start()

    def addClusterQueen(self, queenhost):
        '''
        Inform the ClusterServer about the presence of a ClusterQueen instance
        on the given host.  When the ClusterServer begins to announce work,
        he will do so in "infrastructure mode" and ask any set queens for help.
        '''
        queen = cobra.CobraProxy('cobra://%s:%d/ClusterQueen' % (queenhost, queen_port))
        self.queens.append( queen )

    def shareFileToWorkers(self, filename):
        '''
        Add a file to the list of files which are "shared" to worker clients.
        This allows workers to access a file from the server.

        Example:
            s.shareFileToWorkers('/path/to/file')

        NOTE: Workers may use the openSharedFile() API to access them.
        '''
        self.sharedfiles[filename] = True

    def openSharedFile(self, filename):
        '''
        Return a URI for an open file decriptor for the specified filename.

        NOTE: use openSharedFile() method on work unit to get back a proxy.
        '''
        if not self.sharedfiles.get(filename):
            raise Exception('File %s is not shared!')

        fd = open(filename, 'rb')

        cname = self.cobrad.shareObject(fd, doref=True)
        host,port = cobra.getLocalInfo()

        uri = 'cobra://%s:%d/%s' % (host, port, cname)
        return uri

    def getServerInfo(self): 
        '''
        Return server host/port information
        '''
        return cobra.getLocalInfo() 

    def getServerObjectUri(self, objname):
        '''
        Returns a Cobra URI to access a given object by name.
        Useful for ClusterWork objects to access shared resources.
        '''
        host, port = self.getServerInfo()
        return 'cobra://%s:%d/%s' % (host, port, objname)

    def __touchWork(self, workid):
        # Used to both validate an inprog workid *and*
        # update it's timestamp for the timeout thread
        work = self.inprog.get(workid, None)
        if work is None:
            raise InvalidInProgWorkId(workid)
        work.touch()

    def __cleanWork(self, workid):
        # Used by done/timeout/etc to clea up an in
        # progress work unit
        return self.inprog.pop(workid, None)

    def timerThread(self):
        # Internal function to monitor work unit time
        while self.go:
            try:
                for id,work in self.inprog.items():
                    if work.isTimedOut():
                        self.timeoutWork(work)

            except Exception as e:
                logger.info("ClusterTimer: %s", e)

            time.sleep(2)

    def shutdownServer(self):
        self.go = False

    def announceWork(self):
        """
        Announce to our multicast cluster peers that we have work
        to do! (Or use a queen to proxy to them...)
        """
        if self.queens:
            for q in self.queens:
                try:
                    q.proxyAnnounceWork(self.name, self.cobraname, self.cobrad.port)
                except Exception as e:
                    logger.warning('Queen Error: %s', e)

        else:
            buf = "cobra:%s:%s:%d" % (self.name, self.cobraname, self.cobrad.port)
            self.sendsock.sendto(buf.encode('utf-8'), (cluster_ip, cluster_port))

    def runServer(self, firethread=False):

        if firethread:
            thr = threading.Thread(target=self.runServer, daemon=True)
            thr.start()

        else:
            self.cobrad.fireThread()
            while self.go:

                if len(self.queue):
                    self.announceWork()

                time.sleep(2)

    def inQueueCount(self):
        """
        How long is the current work unit queue.
        """
        return len(self.queue)

    def inProgressCount(self):
        """
        How many work units are in progress?
        """
        return len(self.inprog)

    def addWork(self, work):
        """
        Add a work object to the ClusterServer.  This 
        """
        if not isinstance(work, ClusterWork):
            raise Exception("%s is not a ClusterWork extension!")

        # If this work has no ID, give it one
        if work.id is None:
            work.id = next(self.widiter)

        self.qcond.acquire()
        if self.maxsize is not None:
            while len(self.queue) >= self.maxsize:
                self.qcond.wait()
        self.queue.append(work)
        self.qcond.release()

        if self.callback:
            self.callback.workAdded(self, work)

    def getWork(self):

        self.qcond.acquire()

        try:
            ret = self.queue.popleft()
        except IndexError:
            self.qcond.release()
            return None

        self.qcond.notifyAll()
        self.qcond.release()

        self.inprog[ret.id] = ret
        self.__touchWork(ret.id)

        if self.callback:
            self.callback.workGotten(self, ret)

        return ret


    def doneWork(self, work):
        """
        Used by the clients to report work as done.
        """
        self.__cleanWork(work.id)

        work.done()
        if self.callback:
            self.callback.workDone(self, work)

    def timeoutWork(self, work):
        """
        This method may be over-ridden to handle
        work units that time our for whatever reason.
        """
        self.__cleanWork(work.id)
        if self.callback:
            self.callback.workTimeout(self, work)

    def failWork(self, work):
        """
        This is called for a work unit that is in a failed state.  This is most
        commonly that the work() method has raised an exception.
        """
        self.__cleanWork(work.id)
        if self.callback:
            self.callback.workFailed(self, work)

    def cancelAllWork(self, inprog=True):
        """
        Cancel all of the currently pending work units.  You may
        specify inprog=False to cancel all *queued* work units
        but allow inprogress work units to complete.
        """
        self.qcond.acquire()
        qlist = list(self.queue)
        self.queue.clear()

        if inprog:
            p = self.inprog
            self.inprog = {}
            qlist.extend(p.values())

        self.qcond.notifyAll()
        self.qcond.release()

        if self.callback:
            for w in qlist:
                self.callback.workCanceled(self, w)
        
    def cancelWork(self, workid):
        """
        Cancel a work unit by ID.
        """
        cwork = self.__cleanWork(workid)

        # Remove it from the work queue
        # (if we didn't find in inprog)
        if cwork is None:
            self.qcond.acquire()
            qlist = list(self.queue)
            self.queue.clear()
            for work in qlist:
                if work.id != workid:
                    self.queue.append(work)
                else:
                    cwork = work

            self.qcond.notifyAll()
            self.qcond.release()

        if cwork is None:
            return

        if self.callback:
            self.callback.workCanceled(self, cwork)

    def setWorkStatus(self, workid, status):
        """
        Set the humon readable status for the given work unit.
        """
        self.__touchWork(workid)
        if self.callback:
            self.callback.workStatus(self, workid, status)

    def setWorkCompletion(self, workid, percent):
        """
        Set the percentage completion status for this work unit.
        """
        self.__touchWork(workid)
        if self.callback:
            self.callback.workCompletion(self, workid, percent)

class ClusterClient:

    """
    Listen for our name (or any name if name=="*") on the cobra cluster
    multicast address and if we find a server in need, go help.

    maxwidth is the number of work units to do in parallel
    docode will enable code sharing with the server
    """

    def __init__(self, name, maxwidth=multiprocessing.cpu_count(), docode=False):
        self.go = True
        self.name = name
        self.width = 0
        self.maxwidth = maxwidth
        self.docode = docode

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind(("",cluster_port))
        mreq = struct.pack(b"4sL", socket.inet_aton(cluster_ip), socket.INADDR_ANY)
        self.sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

    def processWork(self):
        """
        Runs handing out work up to maxwidth until self.go == False.
        """
        while self.go:
            
            buf, sockaddr = self.sock.recvfrom(4096)
            if self.width >= self.maxwidth:
                continue

            # make it a string again..
            buf = buf.decode('utf-8')

            server, svrport = sockaddr

            if not buf.startswith("cobra"):
                continue

            info = buf.split(":")

            ilen = len(info)
            if ilen == 4:
                cc,name,cobject,portstr = info
            elif ilen == 5:
                cc,name,cobject,portstr,server = info
            else:
                continue

            if (self.name != name) and (self.name != "*"):
                logger.debug("skipping work, not me...(%r != %r)", name, self.name)
                continue

            port = int(portstr)

            uri = "%s://%s:%d/%s" % (cc,server,port,cobject)
            self.fireRunner(uri)

    def fireRunner(self, uri):
        thr = threading.Thread(target=self.threadForker, args=(uri,), daemon=True)
        thr.start()

    def threadForker(self, uri):
        self.width += 1
        cmd = sub_cmd % (uri, self.docode)
        try:
            sub = subprocess.Popen([sys.executable, '-c', cmd], stdin=subprocess.PIPE)
            sub.wait()
        finally:
            self.width -= 1

class ClusterQueen:

    def __init__(self, ifip, recast=True):

        # Setup our transmission socket
        self.sendsock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sendsock.bind((ifip, 0))

        # FIXME TODO make her optionally a multicast listener that re-broadcasts
        # to her subjects...

    def proxyAnnounceWork(self, name, cobraname, port):
        """
        Send out a multicast announcement to our subjects to go help
        a cluster server.
        """
        # Get the host IP from the connection information
        host, x = cobra.getCallerInfo()
        buf = "cobra:%s:%s:%d:%s" % (name, cobraname, port, host)
        self.sendsock.sendto(buf.encode('utf-8'), (cluster_ip, cluster_port))

def getHostPortFromUri(uri):
    """
    Take the server URI and pull out the
    host and port for use in building the
    dcode uri.
    """
    x = url_req.Request(uri)
    port = None
    hparts = x.host.split(":")
    host = hparts[0]
    if len(hparts):
        port = int(hparts[1])
    return host,port

def workThread(server, work):
    try:
        work.server = server
        work.starttime = time.time()
        work.touch()
        work.work()
        work.endtime = time.time()
        work.server = None
        server.doneWork(work)

    except InvalidInProgWorkId: # the work was canceled
        pass # Nothing to do, the server already knows

    except Exception:
        # Tell the server that the work unit failed
        formatted = traceback.format_exc()
        work.excinfo = formatted
        logger.error(formatted)
        server.failWork(work)

def runAndWaitWork(server, work):

    work.touch()
    thr = threading.Thread(target=workThread, args=(server, work), daemon=True)
    thr.start()

    # Wait around for done or timeout
    while True:
        if work.isTimedOut():
            break

        # If the thread is done, lets get out.
        if not thr.isAlive():
            break

        # If our parent, or some thread closes stdin,
        # time to pack up and go.
        if sys.stdin.closed:
            break

        time.sleep(2)

def getAndDoWork(uri, docode=False):

    # If we wanna use dcode, set it up
    logger.debug("getAndDoWork: uri=", uri)
    try:
        if docode:
            host,port = getHostPortFromUri(uri)
            cobra.dcode.addDcodeServer(host, port=port)

        # Use a cobra proxy with timeout/maxretry so we
        # don't hang forever if the server goes away
        proxy = cobra.CobraProxy(uri, timeout=60, retrymax=3)

        work = proxy.getWork()
        # If we got work, do it.
        if work is not None:
            runAndWaitWork(proxy, work)

    except Exception:
        logger.error(traceback.format_exc())

    # Any way it goes we wanna exit now.  Work units may have
    # spun up non-daemon threads, so lets GTFO.
    gc.collect() # Try to call destructors
    sys.exit(0)  # GTFO




import unittest
import envi.registers as e_regs
import envi.archs.i386.regs as e_i386_regs


class EnviRegisterTest(unittest.TestCase):
    def test_envi_register_slicing(self):
        # we'll use the good ol' i386 registers for demonstrating metaregister slicing
        rctx = e_i386_regs.i386RegisterContext()
        rctx.setRegisterByName('ecx', 0x47145)                                                                                                                                                                     
        self.assertEqual

        self.assertEqual(rctx.getRegisterByName('ecx'), 0x47145)
        self.assertEqual(rctx.getRegisterByName('cx'), 0x7145)
        self.assertEqual(rctx.getRegisterByName('cl'), 0x45)
        self.assertEqual(rctx.getRegisterByName('ch'), 0x71)

        # write-slicing, oversized data
        rctx.setRegisterByName('ch', 0x47145)   # should chop the upper bits out to the correct size
        self.assertEqual(rctx.getRegisterByName('ecx'), 0x44545)



"""
Amd64 Support Module
"""
# Copyright (C) 2007 Invisigoth - See LICENSE file for details
import struct

import envi.archs.amd64 as e_amd64
import vtrace.archs.i386 as vt_i386

class Amd64Mixin(
        e_amd64.Amd64Module,
        e_amd64.Amd64RegisterContext,
        vt_i386.i386WatchMixin):
    """
    Do what we need to for the lucious amd64
    """
    def __init__(self):
        e_amd64.Amd64Module.__init__(self)
        e_amd64.Amd64RegisterContext.__init__(self)
        vt_i386.i386WatchMixin.__init__(self)

        self.setMeta('Architecture','amd64')

    def archGetStackTrace(self):
        self.requireAttached()
        current = 0
        sanity = 1000
        frames = []
        rbp = self.getRegisterByName("rbp")
        rip = self.getRegisterByName("rip")
        frames.append((rip,rbp))

        while rbp != 0 and current < sanity:
            try:
                rbp,rip = self.readMemoryFormat(rbp, "<QQ")
            except:
                break
            frames.append((rip,rbp))
            current += 1

        return frames




'''
Various vtrace exceptions
'''
import ctypes


class PtraceException(Exception):
    def __init__(self, request):
        Exception.__init__(self, 'Ptrace raised exception on request %s (ERRNO: %d)' % (request, ctypes.get_errno()))

class RegisterException(Exception):
    def __init__(self, msg, emuregs, traceregs):
        Exception.__init__(self, msg)
        self.emuregs = emuregs
        self.traceregs = traceregs

class MemoryException(Exception):
    pass

class TargetAddrCalcException(Exception):
    pass




# APIs for Windows 32-bit libraries shell_32.
# Format:  rettype, retname, callconv, exactname, arglist(type, name)
#          arglist type is one of ['int', 'void *']
#          arglist name is one of [None, 'funcptr', 'obj', 'ptr']

api_defs = {
    'shell32.commandlinetoargvw':( 'int', None, 'stdcall', 'shell32.CommandLineToArgW', (('void *', 'ptr'), ('int', None)) ),
    'shell32.shellexecutea':( 'int', None, 'stdcall', 'shell32.ShellExecuteA', (('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'shell32.shellexecutew':( 'int', None, 'stdcall', 'shell32.ShellExecuteW', (('int', None), ('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'shell32.shellexecuteexa':( 'int', None, 'stdcall', 'shell32.ShellExecuteExA', (('void *', 'ptr'),) ),
    'shell32.shellexecuteexw':( 'int', None, 'stdcall', 'shell32.ShellExecuteExW', (('void *', 'ptr'),) ),
    'shell32.shfileoperationa':( 'int', None, 'stdcall', 'shell32.SHFileOperationA', (('void *', 'ptr'),) ),
    'shell32.shfileoperationw':( 'int', None, 'stdcall', 'shell32.SHFileOperationW', (('void *', 'ptr'),) ),
    'shell32.shgetfolderpatha':( 'int', None, 'stdcall', 'shell32.SHGetFolderPathA', (('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'shell32.shgetfolderpathw':( 'int', None, 'stdcall', 'shell32.SHGetFolderPathW', (('int', None), ('int', None), ('int', None), ('int', None), ('void *', 'ptr')) ),
    'shell32.shgetspecialfolderpatha':( 'int', None, 'stdcall', 'shell32.SHGetSpecialFolderPathA', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    'shell32.shgetspecialfolderpathw':( 'int', None, 'stdcall', 'shell32.SHGetSpecialFolderPathW', (('int', None), ('void *', 'ptr'), ('int', None), ('int', None)) ),
    }



import socket

import vstruct
from vstruct.primitives import *
import vstruct.defs.inet as vs_inet

# only the currently parsed record types
DNS_TYPE_A     = 1
DNS_TYPE_NS    = 2
DNS_TYPE_CNAME = 5
DNS_TYPE_SOA   = 6
DNS_TYPE_PTR   = 12
DNS_TYPE_MX    = 15
DNS_TYPE_TXT   = 16
DNS_TYPE_RP    = 17
DNS_TYPE_AFSDB = 18
DNS_TYPE_SIG   = 24
DNS_TYPE_KEY   = 25
DNS_TYPE_AAAA  = 28
DNS_TYPE_LOC   = 29
DNS_TYPE_SRV   = 33
DNS_TYPE_NAPTR = 35
DNS_TYPE_KX    = 36
DNS_TYPE_CERT  = 37
DNS_TYPE_DNAME = 39

DNS_TYPE_OPT   = 41
DNS_TYPE_APL   = 42
DNS_TYPE_DS    = 43
DNS_TYPE_SSHFP = 44
DNS_TYPE_IPSECKEY = 45
DNS_TYPE_RRSIG = 46
DNS_TYPE_NSEC  = 47
DNS_TYPE_DNSKEY= 48
DNS_TYPE_DHCID = 49

DNS_TYPE_NSEC3 = 50
DNS_TYPE_NSEC3PARAM = 51
DNS_TYPE_TLSA  = 52
DNS_TYPE_HIP   = 55
DNS_TYPE_CDS   = 59
DNS_TYPE_CDNSKEY = 60

DNS_TYPE_TKEY  = 249
DNS_TYPE_TSIG  = 250
DNS_TYPE_IXFR  = 251
DNS_TYPE_AXFR  = 252
DNS_TYPE_ANY   = 255
DNS_TYPE_URI   = 256
DNS_TYPE_CAA   = 257

DNS_TYPE_TA    = 32768
DNS_TYPE_DLV   = 32769

dns_type_names = {
    DNS_TYPE_A:     'A',
    DNS_TYPE_NS:    'NS',
    DNS_TYPE_CNAME: 'CNAME',
    DNS_TYPE_SOA:   'SOA',
    DNS_TYPE_PTR:   'PTR',
    DNS_TYPE_MX:    'MX',
    DNS_TYPE_TXT:   'TXT',
    DNS_TYPE_RP    : 'RP',
    DNS_TYPE_AFSDB : 'AFSDB',
    DNS_TYPE_SIG   : 'SIG',
    DNS_TYPE_KEY   : 'KEY',
    DNS_TYPE_AAAA  : 'AAAA',
    DNS_TYPE_LOC   : 'LOC',
    DNS_TYPE_SRV   : 'SRV',
    DNS_TYPE_NAPTR : 'NAPTR',
    DNS_TYPE_KX    : 'KX',
    DNS_TYPE_CERT  : 'CERT',
    DNS_TYPE_DNAME : 'DNAME',
    DNS_TYPE_OPT   : 'OPT',
    DNS_TYPE_APL   : 'APL',
    DNS_TYPE_DS    : 'DS',
    DNS_TYPE_SSHFP : 'SSHFP',
    DNS_TYPE_IPSECKEY : 'IPSECKEY',
    DNS_TYPE_RRSIG : 'RRSIG',
    DNS_TYPE_NSEC  : 'NSEC',
    DNS_TYPE_DNSKEY: 'DNSKEY',
    DNS_TYPE_DHCID : 'DHCID',
    DNS_TYPE_NSEC3 : 'NSEC3',
    DNS_TYPE_NSEC3PARAM : 'NSEC3PARAM',
    DNS_TYPE_TLSA  : 'TLSA',
    DNS_TYPE_HIP   : 'HIP',
    DNS_TYPE_CDS   : 'CDS',
    DNS_TYPE_CDNSKEY : 'CDNSKEY',
    DNS_TYPE_TKEY  : 'TKEY',
    DNS_TYPE_TSIG  : 'TSIG',
    DNS_TYPE_IXFR  : 'IXFR',
    DNS_TYPE_AXFR  : 'AXFR',
    DNS_TYPE_ANY   : 'ANY',
    DNS_TYPE_URI   : 'URI',
    DNS_TYPE_CAA   : 'CAA',
    DNS_TYPE_TA    : 'TA',
    DNS_TYPE_DLV   : 'DLV',
}

DNS_CLASS_IN     = 1
DNS_CLASS_CSNET  = 2
DNS_CLASS_CHAOS  = 3
DNS_CLASS_HESIOD = 4
DNS_CLASS_ANY    = 255

dns_class_names = {
    DNS_CLASS_IN     : 'IN',
    DNS_CLASS_CSNET  : 'CSNET',
    DNS_CLASS_CHAOS  : 'CHAOS',
    DNS_CLASS_HESIOD : 'HESIOD',
    DNS_CLASS_ANY    : 'ANY',
}

DNS_NAMETYPE_LABEL        = 0
DNS_NAMETYPE_RESERVED     = 1
DNS_NAMETYPE_EXTENDED     = 2 
DNS_NAMETYPE_POINTER      = 3
DNS_NAMETYPE_LABELPOINTER = 4

DNS_FLAG_RESP   = 1 << 15
DNS_FLAG_TRUNC  = 1 << 9
DNS_FLAG_RECUR  = 1 << 8
#DNS_FLAG_AA = 1 << 5    # authoritative answer
#DNS_FLAG_TC = 1 << 6    # truncated response
#DNS_FLAG_RD = 1 << 7    # recursion desired
#DNS_FLAG_RA = 1 << 8    # recursion allowed
#DNS_FLAG_AD = 1 << 10   # authentic data
#DNS_FLAG_CD = 1 << 11   # checking disabled

#totally arbitrary count value to abort parsing dns records
DNS_SUSPICIOUS_COUNT = 0x20

class DnsParseError(Exception):
    pass

def dnsFlagsOp(val):
    return (val >> 11) & 0xf

class DnsNameLabel(vstruct.VStruct):
    '''
    A DNS Name component.
    '''
    def __init__(self, label=''):
        vstruct.VStruct.__init__(self)
        self._nametype = 0
        self._pointer  = 0
        self.length = v_uint8( len(label) )
        self.label  = v_bytes( vbytes=label )

    def pcb_length(self):
        size = self.length
        labeltype = size >> 6
        # ordered by use
        if labeltype == 0b11:
            self._nametype = DNS_NAMETYPE_POINTER
            size = 1
        elif labeltype == 0b00:
            self._nametype = DNS_NAMETYPE_LABEL
        elif labeltype == 0b10:
            raise DnsParseError('Extended labeltype is not supported.') 
        elif labeltype == 0b01:
            raise DnsParseError('Unrecognized labeltype (reserved).')
        self.vsGetField('label').vsSetLength(size)

    def pcb_label(self):
        if self.length == 0:
            return
        if self._nametype == DNS_NAMETYPE_POINTER:
            ordlabel = 0
            # some broken(?) packets have a pointer with an empty label
            if self.label:
                ordlabel = ord(self.label)
            # the length field's lower 6 bits + the 8 bits from the label form the pointer
            self._pointer = ((self.length ^ 0xc0) << 8) + ordlabel

    def getNameType(self):
        return self._nametype

    def getNamePointer(self):
        return self._pointer

    def isNamePointer(self):
        if self._nametype == DNS_NAMETYPE_POINTER:
            return True
        return False

    def isNameTerm(self):
        if self.length == 0 or self._nametype == DNS_NAMETYPE_POINTER:
            return True
        return False

class DnsName(vstruct.VArray):
    '''
    The contiguous labels (DnsNameLabel()) in a DNS Name field.  Note that the
    last label may simply be a pointer to an offset earlier in the DNS message.
    '''
    def __init__(self, name=None):
        vstruct.VArray.__init__(self)
        if name is not None:
            for part in name.split('.'):
                self.vsAddElement( DnsNameLabel( part ) )
            self.vsAddElement( DnsNameLabel('') )

    def getTypeVal(self):
        '''
        Return a (nametype, nameval) tuple based on walking the labels.
        '''
        nametype = None
        namepointer = None
        labels = []
        for fname,fobj in self.vsGetFields():
            nametype = fobj.getNameType()
            if nametype == DNS_NAMETYPE_LABEL and fobj.length != 0:
                labels.append(fobj.label)
            if nametype == DNS_NAMETYPE_POINTER:
                namepointer = fobj.getNamePointer()
                if labels:
                    nametype = DNS_NAMETYPE_LABELPOINTER

        joinedlabels = '.'.join(labels)
        if nametype == DNS_NAMETYPE_LABEL:
            return nametype,joinedlabels
        elif nametype == DNS_NAMETYPE_POINTER:
            return nametype,namepointer
        elif nametype == DNS_NAMETYPE_LABELPOINTER:
            return nametype,(joinedlabels,namepointer)
        raise DnsParseError('Unrecognized label.')

    def vsParse(self, bytez, offset=0):
        while offset < len(bytez):
            nl = DnsNameLabel()
            labelofs = offset
            offset = nl.vsParse(bytez, offset=offset)
            self.vsAddElement(nl)
            if nl.isNamePointer() and nl.getNamePointer() >= labelofs:
                raise DnsParseError('Label points forward (or to self).')
            if nl.isNameTerm():
                break
        return offset

class DnsMailboxAsName(DnsName):
    '''
    A DNS Name used to encode a mailbox address.
    '''
    pass

class DnsQuestion(vstruct.VStruct):
    '''
    A DNS Question Record (the query).
    '''
    def __init__(self, name=None, qtype=0, qclass=0):
        vstruct.VStruct.__init__(self)
        self.qname  = DnsName(name=name)
        self.qtype  = v_uint16(qtype, bigend=True)
        self.qclass = v_uint16(qclass, bigend=True)

class DnsQuestionArray(vstruct.VArray):
    '''
    A DNS Question Section.
    '''
    def __init__(self, reccnt):
        vstruct.VArray.__init__(self)
        for i in range(reccnt):
            self.vsAddElement(DnsQuestion())

class DnsResourceRecord(vstruct.VStruct):
    '''
    A DNS Resource Record.  Used in the Answer, Authority, and Additional Sections.
    '''
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.dnsname  = DnsName()
        self.rrtype   = v_uint16(bigend=True)
        self.dnsclass = v_uint16(bigend=True)
        self.ttl      = v_uint32(bigend=True)
        self.rdlength = v_uint16(bigend=True)
        self.rdata    = vstruct.VStruct()

    def pcb_rrtype(self):
        if self.rrtype == DNS_TYPE_A:
            self.rdata.address = vs_inet.IPv4Address()
        elif self.rrtype == DNS_TYPE_AAAA:
            self.rdata.address = vs_inet.IPv6Address()
        elif self.rrtype == DNS_TYPE_NS:
            self.rdata.nsdname = DnsName()
        elif self.rrtype == DNS_TYPE_CNAME:
            self.rdata.cname = DnsName()
        elif self.rrtype == DNS_TYPE_SOA:
            self.rdata.mname   = DnsName()
            # this is an encoded email address 
            self.rdata.rname   = DnsMailboxAsName()
            self.rdata.serial  = v_uint32(bigend=True)
            self.rdata.refresh = v_uint32(bigend=True)
            self.rdata.retry   = v_uint32(bigend=True)
            self.rdata.expire  = v_uint32(bigend=True)
            self.rdata.minimum = v_uint32(bigend=True)
        elif self.rrtype == DNS_TYPE_PTR:
            self.rdata.ptrdname = DnsName()
        elif self.rrtype == DNS_TYPE_MX:
            self.rdata.preference = v_uint16(bigend=True)
            self.rdata.exchange   = DnsName()
        elif self.rrtype == DNS_TYPE_TXT:
            self.rdata.txtdata = v_str()
        else:
            self.rdata.bytez = v_bytes()

    def pcb_rdlength(self):
        size = self.rdlength
        if self.rdata.vsHasField('bytez'):
            self.rdata.vsGetField('bytez').vsSetLength(size)
        elif self.rdata.vsHasField('txtdata'):
            self.rdata.vsGetField('txtdata').vsSetLength(size)


class DnsResourceRecordArray(vstruct.VArray):
    '''
    A DNS RR Section (Answer, Authority, or Additional).
    '''
    def __init__(self, reccnt):
        vstruct.VArray.__init__(self)
        for i in range(reccnt):
            self.vsAddElement(DnsResourceRecord())

class DnsMessage(vstruct.VStruct):
    '''
    A DNS Message.
    '''
    def __init__(self, tcpdns=False):
        vstruct.VStruct.__init__(self)
        self._tcpdns = tcpdns
        if tcpdns:
            self.length = v_uint16(bigend=True)
        self.transid  = v_uint16(bigend=True)
        self.flags    = v_uint16(bigend=True)
        self.qdcount  = v_uint16(bigend=True)
        self.ancount  = v_uint16(bigend=True)
        self.nscount  = v_uint16(bigend=True)
        self.arcount  = v_uint16(bigend=True)
        self.section  = vstruct.VStruct()
        self.section.question   = DnsQuestionArray(0)
        self.section.answer     = DnsResourceRecordArray(0)
        self.section.authority  = DnsResourceRecordArray(0)
        self.section.additional = DnsResourceRecordArray(0)
        self._nptr = {}  # name pointer cache

        #cached question & answers
        self._cache_qrs = None
        self._cache_ars = None

    def pcb_qdcount(self):
        if self.qdcount > DNS_SUSPICIOUS_COUNT:
            raise RuntimeError('DNS suspicious count threshold hit')
        self.section.question = DnsQuestionArray(self.qdcount)

    def pcb_ancount(self):
        if self.ancount > DNS_SUSPICIOUS_COUNT:
            raise RuntimeError('DNS suspicious count threshold hit')
        self.section.answer = DnsResourceRecordArray(self.ancount)

    def pcb_nscount(self):
        if self.nscount > DNS_SUSPICIOUS_COUNT:
            raise RuntimeError('DNS suspicious count threshold hit')
        self.section.authority = DnsResourceRecordArray(self.nscount)

    def pcb_arcount(self):
        if self.arcount > DNS_SUSPICIOUS_COUNT:
            raise RuntimeError('DNS suspicious count threshold hit')
        self.section.additional = DnsResourceRecordArray(self.arcount)

    def vsParse(self, bytez, offset=0):
        self._cache_qrs = None
        self._cache_ars = None
        self._dns_bytes = bytez
        self._dns_offset = offset
        return vstruct.VStruct.vsParse(self, bytez, offset=offset)

    def _getLabelPointerRef(self, msgofs):
        '''
        Given an offset relative to the beginning of a message, create a
        DnsName() structure based on the data there, and return the results
        of its getTypeVal() method (a (nametype, nameval) tuple).
        '''
        # msgofs is relative to the beginning of the message, not necessarily the stream
        if msgofs not in self._nptr:
            # these are often repeated within a message, so we cache them
            self._nptr[msgofs] = DnsName()
            self._nptr[msgofs].vsParse(self._dns_bytes, self._dns_offset + msgofs)
        return self._nptr[msgofs].getTypeVal()

    def getDnsName(self, nametype, nameval):
        '''
        Given a nametype (one of the DNS_NAMETYPE_* constants) and nameval
        (depending on the type, either an fqdn, pointer, or partial fqdn and
        a pointer), return an fqdn.  This is meant to be called with the
        results from a DnsName() instance's getTypeVal() method.
        '''
        if nametype == DNS_NAMETYPE_LABEL:
            fqdn = nameval

        elif nametype == DNS_NAMETYPE_POINTER:
            offset = nameval
            if self._tcpdns:
                # if we're a TCP packet, the 'length' field is not included in the pointer offset
                offset += 2

            fqdn = self.getDnsName(*self._getLabelPointerRef(offset))

        elif nametype == DNS_NAMETYPE_LABELPOINTER:
            beginlabels,offset = nameval
            if self._tcpdns:
                # if we're a TCP packet, the 'length' field is not included in the pointer offset
                offset += 2

            ptrtype,ptrval = self._getLabelPointerRef(offset)
            endlabels = self.getDnsName(ptrtype, ptrval)
            fqdn = '.'.join((beginlabels, endlabels))
        return fqdn

    def getQuestionRecords(self):
        '''
        Return a list of Question records as (dnstype, dnsclass, fqdn) tuples.
        '''
        if self._cache_qrs:
            return self._cache_qrs
        ret = []
        for fname,q in self.section.question.vsGetFields():
            fqdn = self.getDnsName(*q.qname.getTypeVal())
            ret.append((q.qtype, q.qclass, fqdn))
        self._cache_qrs = ret
        return ret

    def _getResourceRecords(self, structure):
        '''
        Given a DnsResourceRecordArray() structure, return a list of Resource 
        Records as (dnstype, dnsclass, ttl, fqdn, adata) tuples.  If a parser
        is available for the dnsclass, the 'rdata' field will be further parsed 
        into its components (as a tuple if necessary).
        '''
        ret = []
        for fname,rr in structure.vsGetFields():
            fqdn = self.getDnsName(*rr.dnsname.getTypeVal())

            rdata = None
            if rr.rrtype == DNS_TYPE_A:
                rdata = vs_inet.reprIPv4Addr(rr.rdata.address)
            elif rr.rrtype == DNS_TYPE_AAAA:
                rdata = vs_inet.reprIPv6Addr(rr.rdata.address)
            elif rr.rrtype == DNS_TYPE_NS:
                rdata = self.getDnsName(*rr.rdata.nsdname.getTypeVal())
            elif rr.rrtype == DNS_TYPE_CNAME:
                rdata = self.getDnsName(*rr.rdata.cname.getTypeVal())
            elif rr.rrtype == DNS_TYPE_SOA:
                rdata = (self.getDnsName(*rr.rdata.mname.getTypeVal()),
                         self.getDnsName(*rr.rdata.rname.getTypeVal()),
                         rr.rdata.serial,
                         rr.rdata.refresh,
                         rr.rdata.retry, 
                         rr.rdata.expire,
                         rr.rdata.minimum)
            elif rr.rrtype == DNS_TYPE_PTR:
                rdata = self.getDnsName(*rr.rdata.ptrdname.getTypeVal())
            elif rr.rrtype == DNS_TYPE_MX:
                rdata = (rr.rdata.preference,
                         self.getDnsName(*rr.rdata.exchange.getTypeVal()))
            elif rr.rrtype == DNS_TYPE_TXT:
                rdata = rr.rdata.txtdata
            else:
                rdata = rr.rdata.bytez

            ret.append((rr.rrtype, rr.dnsclass, rr.ttl, fqdn, rdata))
        return ret

    def getAnswerRecords(self):
        '''
        Return a list of Answer records as (rrtype, dnsclass, ttl, fqdn,
        rdata) tuples.  If a parser is available for the dnsclass, the 
        'rdata' field will be further parsed into its components (as a
        tuple if necessary).
        '''
        if not self._cache_ars:
            self._cache_ars = self._getResourceRecords(structure=self.section.answer)
        return self._cache_ars

    def getAuthorityRecords(self):
        '''
        Return a list of Authority records as (rrtype, dnsclass, ttl, 
        fqdn, rdata) tuples.  If a parser is available for the dnsclass, 
        the 'rdata' field will be further parsed into its components 
        (as a tuple if necessary).
        '''
        return self._getResourceRecords(structure=self.section.authority)

    def getAdditionalRecords(self):
        '''
        Return a list of Additional records as (rrtype, dnsclass, ttl, 
        fqdn, rdata) tuples.  If a parser is available for the dnsclass, 
        the 'rdata' field will be further parsed into its components (as 
        a tuple if necessary).
        '''
        return self._getResourceRecords(structure=self.section.additional)

    def getDnsNames(self):
        '''
        Return a list of the DNS names in the message.
        '''
        fqdns = set()
        for ofs,indent,fname,fobj in self.vsGetPrintInfo():
            if fobj.vsGetTypeName() == 'DnsName':
                fqdns.add(self.getDnsName(*fobj.getTypeVal()))
        return list(fqdns)

    def getIPv4Integers(self):
        '''
        Return a list of the IPv4 addresses in the message.
        '''
        ips = set()
        for ofs,indent,fname,fobj in self.vsGetPrintInfo():
            if fobj.vsGetTypeName() == 'IPv4Address':
                ips.add(fobj._vs_value)
        return list(ips)

    def getEmailAddresses(self):
        '''
        Return a list of the email addresses which are encoded as DNS names
        in the message (they are decoded back to email addresses here).
        '''
        emails = set()
        for ofs,indent,fname,fobj in self.vsGetPrintInfo():
            if fobj.vsGetTypeName() == 'DnsMailboxAsName':
                mailbox = self.getDnsName(*fobj.getTypeVal())
                parts = mailbox.split('.', 1)
                emails.add('@'.join(parts))
        return list(emails)



import unittest

import envi.config as e_config


class ConfigTest(unittest.TestCase):
    def test_config_params(self):
        defaults = {
            'woot': 10,
            'biz': '0x41414141',
            'foosub': {
                'bar': 'qwer',
                'baz': ('one', 'two', 'three'),
            }
        }
        cfg = e_config.EnviConfig(defaults=defaults)

        self.assertEqual(cfg.woot, 10)
        self.assertEqual(cfg.foosub.bar, 'qwer')
        self.assertEqual(cfg.foosub.baz[0], 'one')
        self.assertEqual(cfg.biz, '0x41414141')
        for idx, item in enumerate(cfg.foosub.baz):
            self.assertEqual(defaults['foosub']['baz'][idx], item)

    def test_config_by_string(self):
        defaults = {
            'woot': 10,
            'biz': '0x41414141',
            'foosub': {
                'bar': 'qwer',
                'baz': ('one', 'two', 'three'),
            }
        }
        cfg = e_config.EnviConfig(defaults=defaults)

        self.assertEqual(cfg.getOptionByString('woot'), 0xa)
        self.assertEqual(cfg.getOptionByString('biz'), '0x41414141')
        self.assertEqual(cfg.getOptionByString('foosub.bar'), 'qwer')
        self.assertEqual(cfg.getOptionByString('foosub.baz'), ('one', 'two', 'three'))

